import traceback
import uuid
from collections import defaultdict
from multiprocessing.pool import ThreadPool
from types import FunctionType
from typing import Any, Callable

import pandas as pd

from qubx import logger
from qubx.core.basics import (
    DataType,
    Deal,
    FundingPayment,
    InitializingSignal,
    Instrument,
    MarketEvent,
    Order,
    RestoredState,
    Signal,
    TargetPosition,
    Timestamped,
    TriggerEvent,
    dt_64,
    td_64,
)
from qubx.core.errors import BaseErrorEvent
from qubx.core.exceptions import StrategyExceededMaxNumberOfRuntimeFailuresError
from qubx.core.helpers import BasicScheduler, CachedMarketDataHolder, process_schedule_spec
from qubx.core.interfaces import (
    IAccountProcessor,
    IHealthMonitor,
    IMarketManager,
    IPositionGathering,
    IProcessingManager,
    IStrategy,
    IStrategyContext,
    ISubscriptionManager,
    ITimeProvider,
    ITradeDataExport,
    IUniverseManager,
    PositionsTracker,
)
from qubx.core.loggers import StrategyLogging
from qubx.core.series import Bar, OrderBook, Quote, Trade
from qubx.core.stale_data_detector import StaleDataDetector
from qubx.trackers.riskctrl import _InitializationStageTracker


class ProcessingManager(IProcessingManager):
    MAX_NUMBER_OF_STRATEGY_FAILURES: int = 10
    DATA_READY_TIMEOUT: td_64 = td_64(60, "s")

    _context: IStrategyContext
    _strategy: IStrategy
    _logging: StrategyLogging
    _market_data: IMarketManager
    _subscription_manager: ISubscriptionManager
    _time_provider: ITimeProvider
    _account: IAccountProcessor
    _position_tracker: PositionsTracker
    _position_gathering: IPositionGathering
    _warmup_position_gathering: IPositionGathering
    _cache: CachedMarketDataHolder
    _scheduler: BasicScheduler
    _universe_manager: IUniverseManager
    _exporter: ITradeDataExport | None = None
    _health_monitor: IHealthMonitor
    _stale_data_detector: StaleDataDetector

    _handlers: dict[str, Callable[["ProcessingManager", Instrument, str, Any], TriggerEvent | None]]
    _strategy_name: str

    _trigger_on_time_event: bool = False
    _fit_is_running: bool = False
    _fails_counter: int = 0
    _is_simulation: bool
    _pool: ThreadPool | None
    _trig_bar_freq_nsec: int | None = None
    _cur_sim_step: int | None = None
    _updated_instruments: set[Instrument] = set()
    _data_ready_start_time: dt_64 | None = None
    _last_data_ready_log_time: dt_64 | None = None
    _all_instruments_ready_logged: bool = False
    _emitted_signals: list[Signal] = []  # signals that were emitted
    _active_targets: dict[Instrument, TargetPosition] = {}

    # - post-warmup initialization
    _init_stage_position_tracker: PositionsTracker
    _instruments_in_init_stage: set[Instrument] = set()

    # - custom scheduled methods
    _custom_scheduled_methods: dict[str, Callable] = {}

    def __init__(
        self,
        context: IStrategyContext,
        strategy: IStrategy,
        logging: StrategyLogging,
        market_data: IMarketManager,
        subscription_manager: ISubscriptionManager,
        time_provider: ITimeProvider,
        account: IAccountProcessor,
        position_tracker: PositionsTracker,
        position_gathering: IPositionGathering,
        warmup_position_gathering: IPositionGathering,
        universe_manager: IUniverseManager,
        cache: CachedMarketDataHolder,
        scheduler: BasicScheduler,
        is_simulation: bool,
        health_monitor: IHealthMonitor,
        exporter: ITradeDataExport | None = None,
    ):
        self._context = context
        self._strategy = strategy
        self._logging = logging
        self._market_data = market_data
        self._subscription_manager = subscription_manager
        self._time_provider = time_provider
        self._account = account
        self._is_simulation = is_simulation
        self._position_gathering = position_gathering
        self._position_tracker = position_tracker
        self._universe_manager = universe_manager
        self._cache = cache
        self._scheduler = scheduler
        self._exporter = exporter
        self._health_monitor = health_monitor

        # Initialize stale data detector with default disabled state
        # Will be configured later based on strategy settings
        self._stale_data_detector = StaleDataDetector(
            cache=cache,
            time_provider=time_provider,
        )
        self._stale_data_detection_enabled = False

        self._pool = ThreadPool(2) if not self._is_simulation else None
        self._handlers = {
            n.split("_handle_")[1]: f
            for n, f in self.__class__.__dict__.items()
            if type(f) is FunctionType and n.startswith("_handle_")
        }
        self._strategy_name = strategy.__class__.__name__
        self._trig_bar_freq_nsec = None
        self._updated_instruments = set()
        self._data_ready_start_time = None
        self._last_data_ready_log_time = None
        self._all_instruments_ready_logged = False
        self._emitted_signals = []

        # - special tracker for post-warmup initialization signals
        self._init_stage_position_tracker = _InitializationStageTracker()
        self._instruments_in_init_stage = set()
        self._active_targets = {}
        self._custom_scheduled_methods = {}

        self._warmup_position_gathering = warmup_position_gathering

        # - schedule daily delisting check at 23:30 (end of day)
        self._scheduler.schedule_event("30 23 * * *", "delisting_check")

    def set_fit_schedule(self, schedule: str) -> None:
        logger.info(f"Setting fit schedule to {schedule}")
        rule = process_schedule_spec(schedule)
        if rule.get("type") != "cron":
            raise ValueError("Only cron type is supported for fit schedule")
        self._scheduler.schedule_event(rule["schedule"], "fit")

    def set_event_schedule(self, schedule: str) -> None:
        logger.info(f"Setting event schedule to {schedule}")
        rule = process_schedule_spec(schedule)
        if not rule or "type" not in rule:
            raise ValueError(f"Can't recognize schedule format: '{schedule}'")

        if rule["type"] != "cron":
            raise ValueError("Only cron type is supported for event schedule")

        # Unschedule existing time event first to avoid conflicts
        self._scheduler.unschedule_event("time")

        self._scheduler.schedule_event(rule["schedule"], "time")
        self._trigger_on_time_event = True

    def get_event_schedule(self, event_id: str) -> str | None:
        return self._scheduler.get_schedule_for_event(event_id)

    def schedule(self, cron_schedule: str, method: Callable[["IStrategyContext"], None]) -> None:
        """
        Register a custom method to be called at specified times.

        Args:
            cron_schedule: Cron-like schedule string (e.g., "0 0 * * *" for daily at midnight)
            method: Method to call when schedule triggers
        """
        rule = process_schedule_spec(cron_schedule)
        if not rule or rule.get("type") != "cron":
            raise ValueError("Only cron type is supported for custom schedules")

        # Generate unique event ID for this custom schedule
        event_id = f"custom_schedule_{str(uuid.uuid4()).replace('-', '_')}"

        # Store the method reference
        self._custom_scheduled_methods[event_id] = method

        # Schedule the event
        self._scheduler.schedule_event(rule["schedule"], event_id)

    def configure_stale_data_detection(
        self, enabled: bool, detection_period: str | None = None, check_interval: str | None = None
    ) -> None:
        """
        Configure stale data detection settings.

        Args:
            enabled: Whether to enable stale data detection
            detection_period: Period to consider data as stale (e.g., "5Min", "1h"). If None, uses detector default.
            check_interval: Interval between stale data checks (e.g., "30s", "1Min"). If None, uses detector default.
        """
        self._stale_data_detection_enabled = enabled

        if enabled and (detection_period is not None or check_interval is not None):
            # Recreate the detector with new parameters
            kwargs = {}
            if detection_period is not None:
                kwargs["detection_period"] = detection_period
            if check_interval is not None:
                kwargs["check_interval"] = check_interval

            self._stale_data_detector = StaleDataDetector(
                cache=self._cache, time_provider=self._time_provider, **kwargs
            )

    def process_data(self, instrument: Instrument, d_type: str, data: Any, is_historical: bool) -> bool:
        should_stop = self.__process_data(instrument, d_type, data, is_historical)
        if not is_historical:
            self._logging.notify(self._time_provider.time())
            if self._context.emitter is not None:
                self._context.emitter.notify(self._context)
        return should_stop

    def is_fitted(self) -> bool:
        return self._context._strategy_state.is_on_fit_called

    def __process_data(self, instrument: Instrument, d_type: str, data: Any, is_historical: bool) -> bool:
        handler = self._handlers.get(d_type)
        with self._health_monitor("process_data"):
            if not d_type:
                event = None
            elif is_historical:
                event = self._process_hist_event(instrument, d_type, data)
            elif handler:
                event = handler(self, instrument, d_type, data)
            else:
                event = self._process_custom_event(instrument, d_type, data)

        if not self._context._strategy_state.is_on_start_called and not self._is_order_update(d_type):
            self._handle_start()

        if (
            not self._context._strategy_state.is_on_warmup_finished_called
            and not self._context._strategy_state.is_warmup_in_progress
            and not self._is_order_update(d_type)
        ):
            if self._context.get_warmup_positions() or self._context.get_warmup_orders():
                self._handle_state_resolution()

            # Restore tracker and gatherer state if available
            restored_state = self._context.get_restored_state()
            if restored_state is not None:
                self._restore_tracker_and_gatherer_state(restored_state)

            self._handle_warmup_finished()

        # - check if it still didn't call on_fit() for first time
        if not self._context._strategy_state.is_on_fit_called and not self._fit_is_running:
            self._handle_fit(None, "fit", (None, self._time_provider.time()))
            return False

        if not event and not self._emitted_signals:
            return False

        # - if fit was not called - skip on_event call
        if not self._context._strategy_state.is_on_fit_called or (
            not self._is_simulation and not self._context._strategy_state.is_on_warmup_finished_called
        ):
            return False

        # - if strategy still fitting - skip on_event call
        if self._fit_is_running:
            # logger.debug(
            #     f"Skipping {self._strategy_name}::on_event({instrument}, {d_type}, [...], {is_historical}) fitting in progress (orders and deals processed)!"
            # )
            return False

        signals: list[Signal] = []

        # - if there are signals that were emitted before, we need to process them first
        if self._emitted_signals:
            signals.extend(self._emitted_signals)
            self._emitted_signals.clear()

        try:
            # - some small optimization
            _is_market_ev = isinstance(event, MarketEvent)
            _is_trigger_ev = isinstance(event, TriggerEvent)

            if _is_market_ev:
                with self._health_monitor("stg.market_event"):
                    signals.extend(self._as_list(self._strategy.on_market_data(self._context, event)))

            if _is_trigger_ev or (_is_market_ev and event.is_trigger):
                _trigger_event = event.to_trigger() if _is_market_ev else event

                # FIXME: (2025-06-17) we need to refactor this to avoid doing it here !!!
                # - on trigger event we need to be sure that all instruments have finalized OHLC data
                # - - - - - - IMPORTANT NOTES - - - - - -
                # This is a temporary fix to ensure that all instruments have finalized OHLC data.
                # In live mode with multiple instruments, we can have a situation where one instrument
                # is updated with new bar data while other instruments are not updated yet.
                # This leads to a situation where indicators are not calculated correctly for all instruments in the universe.
                # A simple dirty solution is to update OHLC data for all instruments in the universe with the last update value but with the actual time.
                # This leads to finalization of OHLC data, but the open price may differ slightly from the real one.
                # - - - - - - - - - - - - - - - - - - - -

                # - finalize OHLC data for all instruments
                self._cache.finalize_ohlc_for_instruments(event.time, self._context.instruments)

                with self._health_monitor("stg.trigger_event"):
                    signals.extend(self._as_list(self._strategy.on_event(self._context, _trigger_event)))

                # - we reset failures counter when we successfully process on_event
                self._fails_counter = 0

            if isinstance(event, Order):
                with self._health_monitor("stg.order_update"):
                    signals.extend(self._as_list(self._strategy.on_order_update(self._context, event)))

            self._subscription_manager.commit()  # apply pending operations

        except Exception as strat_error:
            # Record event dropped due to exception
            self._health_monitor.record_event_dropped(d_type)

            # - probably we need some cooldown interval after exception to prevent flooding
            logger.error(f"Strategy {self._strategy_name} raised an exception: {strat_error}")
            logger.opt(colors=False).error(traceback.format_exc())

            #  - we stop execution after maximal number of errors in a row
            self._fails_counter += 1
            if self._fails_counter >= self.MAX_NUMBER_OF_STRATEGY_FAILURES:
                logger.error(f"STRATEGY FAILED {self.MAX_NUMBER_OF_STRATEGY_FAILURES} TIMES IN THE ROW - STOPPING ...")
                raise StrategyExceededMaxNumberOfRuntimeFailuresError()

        # - process and execute signals if they are provided
        if signals:
            with self._health_monitor("position_processing"):
                self.__process_signals(signals)

        return False

    def _get_tracker_for(self, instrument: Instrument) -> PositionsTracker:
        return (
            self._init_stage_position_tracker
            if instrument in self._instruments_in_init_stage
            else self._position_tracker
        )

    def _get_position_gatherer(self) -> IPositionGathering:
        return (
            self._position_gathering
            if self._context._strategy_state.is_on_warmup_finished_called
            else self._warmup_position_gathering
        )

    def __preprocess_signals_and_split_by_stage(
        self, signals: list[Signal]
    ) -> tuple[list[Signal], list[Signal], set[Instrument]]:
        """
        Preprocess signals:
            - split into standard and initializing signals
            - prevent service signals to be processed
            - set strategy group name if not set
            - update reference prices for signals
            - switch tracker for initializing signals to the post-warmup one
            - return list of standard signals, list of initializing signals, and set of instruments to cancel post-warmup trackers for
        """
        _init_signals: list[Signal] = []
        _std_signals: list[Signal] = []
        _cancel_init_stage_instruments_tracker = set()

        for signal in signals:
            instr = signal.instrument

            # - set strategy group name if not set
            if not signal.group:
                signal.group = self._strategy_name

            # - update reference prices for signals
            if signal.reference_price is None:
                if q := self._market_data.quote(instr):
                    signal.reference_price = q.mid_price()

            # - prevent service signals to be processed
            if signal.is_service:
                continue

            # - if there is initializing signal, we need to switch tracker to the post-warmup one for this instrument
            if isinstance(signal, InitializingSignal):
                # - if target is already active and it receives initializing signal
                # - it means that something was sent wrong and we need to skip this signal
                if instr in self.get_active_targets():
                    logger.warning(
                        f"Skip initializing signal <y>{signal}</y> for <g>{instr}</g> because strategy has active target {self.get_active_targets()[instr]} !"
                    )
                else:
                    _init_signals.append(signal)
                    self._instruments_in_init_stage.add(instr)
                    logger.debug(f"Switching tracker for <g>{instr}</g> to post-warmup initialization")
            else:
                _std_signals.append(signal)
                if instr in self._instruments_in_init_stage:
                    _cancel_init_stage_instruments_tracker.add(instr)
                    self._instruments_in_init_stage.remove(instr)
                    logger.debug(f"Switching tracker for <g>{instr}</g> back to defined position tracker")

        return _std_signals, _init_signals, _cancel_init_stage_instruments_tracker

    def __process_signals(self, signals: list[Signal]):
        _targets_from_trackers: list[TargetPosition] = []

        # - preprocess signals: split into usual and initializing signals
        _std_signals, _init_signals, _cancel_init_trackers_for = self.__preprocess_signals_and_split_by_stage(signals)

        # - cancel post-warmup trackers
        if _cancel_init_trackers_for:
            for instr in _cancel_init_trackers_for:
                self._init_stage_position_tracker.cancel_tracking(self._context, instr)

        # - notify post-warmup position tracker for the new initializing signals
        if _init_signals:
            _targets_from_trackers.extend(
                self._as_list(self._init_stage_position_tracker.process_signals(self._context, _init_signals))
            )

        # - notify position tracker for the new signals
        if _std_signals:
            _targets_from_trackers.extend(
                self._as_list(self._position_tracker.process_signals(self._context, _std_signals))
            )

        # - notify position gatherer for the new target positions
        if _targets_from_trackers:
            self._get_position_gatherer().alter_positions(
                self._context, self.__preprocess_and_log_target_positions(_targets_from_trackers)
            )

        # - log all signals and export signals if exporter is specified after processing because trackers can modify the signals
        self._logging.save_signals(signals)

        # - export signals if exporter is specified
        if self._exporter is not None and signals:
            self._exporter.export_signals(self._time_provider.time(), signals, self._account)

        # - emit signals to metric emitters if available
        if self._context.emitter is not None and signals:
            self._context.emitter.emit_signals(
                self._time_provider.time(), signals, self._account, _targets_from_trackers
            )

    def __invoke_on_fit(self) -> None:
        with self._health_monitor("ctx.on_fit"):
            try:
                logger.debug(f"[<y>{self.__class__.__name__}</y>] :: Invoking <g>{self._strategy_name}</g> on_fit")
                self._strategy.on_fit(self._context)
                self._subscription_manager.commit()  # apply pending operations
                logger.debug(f"[<y>{self.__class__.__name__}</y>] :: <g>{self._strategy_name}</g> is fitted")
            except Exception as strat_error:
                logger.error(
                    f"[{self.__class__.__name__}] :: Strategy {self._strategy_name} on_fit raised an exception: {strat_error}"
                )
                logger.opt(colors=False).error(traceback.format_exc())
            finally:
                self._fit_is_running = False
                self._context._strategy_state.is_on_fit_called = True

    def __preprocess_and_log_target_positions(self, target_positions: list[TargetPosition]) -> list[TargetPosition]:
        filtered_target_positions = []

        if target_positions:
            # - check if trading is allowed for each target position
            for t in target_positions:
                _instr = t.instrument
                if self._universe_manager.is_trading_allowed(_instr):
                    filtered_target_positions.append(t)

                    # - new position will be non-zero
                    if abs(t.target_position_size) > _instr.min_size:
                        self._active_targets[_instr] = t
                    else:
                        self._active_targets.pop(_instr, None)

            # - log target positions
            self._logging.save_targets(filtered_target_positions)

            # - export target positions if exporter is available
            if self._exporter is not None:
                self._exporter.export_target_positions(self._time_provider.time(), target_positions, self._account)

        return filtered_target_positions

    def _run_in_thread_pool(self, func: Callable, args=()):
        # For simulation we don't need to call function in thread
        if self._is_simulation:
            func(*args)
        else:
            assert self._pool
            self._pool.apply_async(func, args)

    @staticmethod
    def _as_list(xs: list[Any] | Any | None) -> list[Any]:
        if xs is None:
            return []

        if isinstance(xs, list):
            return xs

        return [xs]

    __SUBSCR_TO_DATA_MATCH_TABLE = {
        DataType.OHLC: [Bar],
        DataType.OHLC_QUOTES: [Quote, OrderBook],
        DataType.OHLC_TRADES: [Trade],
        DataType.QUOTE: [Quote],
        DataType.TRADE: [Trade],
        DataType.ORDERBOOK: [OrderBook],
    }

    def _is_base_data(self, data: Timestamped) -> tuple[bool, Timestamped]:
        _base_ss = DataType.from_str(self._subscription_manager.get_base_subscription())[0]
        _d_probe = data
        return (
            type(_d_probe) in _rule if (_rule := self.__SUBSCR_TO_DATA_MATCH_TABLE.get(_base_ss)) else False,
            _d_probe,
        )

    def _is_data_ready(self) -> bool:
        """
        Check if strategy can start based on data availability with timeout logic.

        Two-phase approach:
        - Phase 1 (0-DATA_READY_TIMEOUT): Wait for ALL instruments to have data
        - Phase 2 (after timeout): Wait for at least 1 instrument to have data

        Returns:
            bool: True if strategy can start, False if still waiting
        """
        total_instruments = len(self._context.instruments)

        # Handle edge case: no instruments
        if total_instruments == 0:
            return True

        ready_instruments = len(self._updated_instruments)
        current_time = self._time_provider.time()

        # Record start time on first call
        if self._data_ready_start_time is None:
            self._data_ready_start_time = current_time

        # Phase 1: Try to get all instruments ready within timeout
        elapsed_td = current_time - self._data_ready_start_time

        if elapsed_td <= self.DATA_READY_TIMEOUT:
            # Within timeout period - wait for ALL instruments
            if ready_instruments == total_instruments:
                if not self._all_instruments_ready_logged:
                    logger.info(f"All {total_instruments} instruments have data - strategy ready to start")
                    self._all_instruments_ready_logged = True
                return True
            else:
                # Log periodic status during Phase 1 - throttled to once per 10 seconds
                elapsed_seconds = elapsed_td / td_64(1, "s")
                should_log = self._last_data_ready_log_time is None or (
                    current_time - self._last_data_ready_log_time
                ) >= td_64(10, "s")

                if should_log and elapsed_seconds > 0:
                    missing_instruments = set(self._context.instruments) - self._updated_instruments
                    missing_symbols = [inst.symbol for inst in missing_instruments]
                    remaining_timeout = (self.DATA_READY_TIMEOUT - elapsed_td) / td_64(1, "s")
                    logger.info(
                        f"Waiting for all instruments ({ready_instruments}/{total_instruments} ready). "
                        f"Missing: {missing_symbols}. Will start with partial data in {remaining_timeout:.0f}s"
                    )
                    self._last_data_ready_log_time = current_time
                return False
        else:
            # Phase 2: After timeout - need at least 1 instrument
            if ready_instruments >= total_instruments:
                if not self._all_instruments_ready_logged:
                    logger.info(f"All {total_instruments} instruments have data - strategy ready to start")
                    self._all_instruments_ready_logged = True
                return True

            elif ready_instruments >= 1:
                missing_instruments = set(self._context.instruments) - self._updated_instruments
                missing_symbols = [inst.symbol for inst in missing_instruments]

                # Log once when entering Phase 2
                should_log = self._last_data_ready_log_time is None or (
                    current_time - self._last_data_ready_log_time
                ) >= td_64(10, "s")
                if should_log:
                    logger.info(
                        f"Timeout reached - starting with {ready_instruments}/{total_instruments} instruments ready. "
                        f"Missing: {missing_symbols}"
                    )
                    self._last_data_ready_log_time = current_time
                return True
            else:
                # Still no instruments ready - keep waiting and log periodically
                should_log = self._last_data_ready_log_time is None or (
                    current_time - self._last_data_ready_log_time
                ) >= td_64(10, "s")
                if should_log:
                    # logger.warning(
                    #     f"No instruments ready after timeout - still waiting "
                    #     f"({ready_instruments}/{total_instruments} ready)"
                    # )
                    self._last_data_ready_log_time = current_time
                return False

    def __update_base_data(
        self, instrument: Instrument, event_type: str, data: Timestamped, is_historical: bool = False
    ) -> bool:
        """
        Updates the base data cache with the provided data.

        Returns:
            bool: True if the data is base data and the strategy should be triggered, False otherwise.
        """
        is_base_data, _update = self._is_base_data(data)
        # logger.info(f"{_update} {is_base_data and not self._trigger_on_time_event}")

        # update cached ohlc is this is base subscription
        _update_ohlc = is_base_data
        self._cache.update(
            instrument,
            event_type,
            _update,
            update_ohlc=_update_ohlc,
            is_historical=is_historical,
            is_base_data=is_base_data,
        )

        # update trackers, gatherers on base data
        if not is_historical:
            if is_base_data:
                # - mark instrument as updated
                self._updated_instruments.add(instrument)

                # - update position price
                self._account.update_position_price(self._time_provider.time(), instrument, _update)

                # - update tracker
                _targets_from_tracker = self._get_tracker_for(instrument).update(self._context, instrument, _update)

                # - notify position gatherer for the new target positions
                if _targets_from_tracker:
                    # - tracker generated new targets on update, notify position gatherer
                    self._get_position_gatherer().alter_positions(
                        self._context, self.__preprocess_and_log_target_positions(self._as_list(_targets_from_tracker))
                    )

                # - update position gatherer with market data
                self._get_position_gatherer().update(self._context, instrument, _update)

                # - check for stale data periodically (only for base data updates)
                # This ensures we only check when we have new meaningful data
                if self._stale_data_detection_enabled and self._context._strategy_state.is_on_start_called:
                    stale_instruments = self._stale_data_detector.detect_stale_instruments(self._context.instruments)
                    if stale_instruments:
                        for instr in stale_instruments:
                            logger.info(f"Detected stale data for instrument {instr.symbol}")
                        logger.info(
                            f"Removing {len(stale_instruments)} stale instruments from universe: {[i.symbol for i in stale_instruments]}"
                        )
                        self._universe_manager.remove_instruments(stale_instruments, if_has_position_then="close")
            else:
                # - if it's not base data, we need to process it as market data
                self._account.process_market_data(self._time_provider.time(), instrument, _update)

        return is_base_data and not self._trigger_on_time_event

    ###########################################################################
    # - Handlers for different types of incoming data
    ###########################################################################

    # it's important that we call it with _process to not include in the handlers map
    def _process_custom_event(
        self, instrument: Instrument | None, event_type: str, event_data: Any
    ) -> MarketEvent | None:
        # Handle custom scheduled events
        if event_type in self._custom_scheduled_methods:
            try:
                method = self._custom_scheduled_methods[event_type]
                method(self._context)
                logger.debug(f"[ProcessingManager] :: Executed custom scheduled method for event: {event_type}")
            except Exception as e:
                logger.error(
                    f"[ProcessingManager] :: Error executing custom scheduled method for event {event_type}: {e}"
                )
                logger.opt(colors=False).error(traceback.format_exc())
            # Don't return a MarketEvent for custom scheduled events - they shouldn't trigger strategy.on_event
            return None

        if instrument is not None:
            self.__update_base_data(instrument, event_type, event_data)

        elif instrument is None and isinstance(event_data, dict):
            for _instrument, data in event_data.items():
                if isinstance(_instrument, Instrument):
                    self.__update_base_data(_instrument, event_type, data)

        return MarketEvent(self._time_provider.time(), event_type, instrument, event_data)

    def _process_hist_event(self, instrument: Instrument, event_type: str, event_data: Any) -> None:
        if not isinstance(event_data, list):
            event_data = [event_data]
        if DataType.OHLC == event_type:
            # - update ohlc using the list directly, this allows to update
            # multiple timeframes with different data (1h can have more bars than 1m)
            _, sub_params = DataType.from_str(event_type)
            timeframe = sub_params.get("timeframe", self._cache.default_timeframe)

            self._cache.update_by_bars(instrument, timeframe, event_data)
        else:
            for data in event_data:
                self.__update_base_data(instrument, event_type, data, is_historical=True)

    def _handle_event(self, instrument: Instrument, event_type: str, event_data: Any) -> TriggerEvent:
        return TriggerEvent(self._time_provider.time(), event_type, instrument, event_data)

    def _handle_time(self, instrument: Instrument, event_type: str, data: dt_64) -> TriggerEvent:
        return TriggerEvent(self._time_provider.time(), event_type, instrument, data)

    def _handle_service_time(self, instrument: Instrument, event_type: str, data: dt_64) -> TriggerEvent | None:
        """It is used by simulation as a dummy to trigger actual time events."""
        pass

    def _handle_start(self) -> None:
        if not self._is_data_ready():
            return
        self._strategy.on_start(self._context)
        self._context._strategy_state.is_on_start_called = True

    def _handle_state_resolution(self) -> None:
        if not self._is_data_ready():
            return

        _ctx = self._context
        resolver = _ctx.initializer.get_state_resolver()
        if resolver is None:
            logger.warning("No state resolver found, skipping state resolution")
            return

        self._log_state_mismatch()

        resolver_name = (
            getattr(resolver, "__name__", str(resolver))
            if callable(resolver) and hasattr(resolver, "__name__")
            else resolver.__class__.__name__
        )

        logger.info(f"<yellow>Resolving state mismatch with:</yellow> <g>{resolver_name}</g>")

        resolver(_ctx, _ctx.get_warmup_positions(), _ctx.get_warmup_orders(), _ctx.get_warmup_active_targets())

    def _restore_tracker_and_gatherer_state(self, restored_state: RestoredState) -> None:
        """
        Restore state for position tracker and gatherer.

        Args:
            restored_state: The restored state containing signals and target positions
        """
        if not self._is_data_ready():
            return

        # Restore tracker state from signals
        all_signals = []
        for instrument, signals in restored_state.instrument_to_signal_positions.items():
            all_signals.extend(signals)

        if all_signals:
            logger.info(f"<yellow>Restoring tracker state from {len(all_signals)} signals</yellow>")
            self._position_tracker.restore_position_from_signals(self._context, all_signals)

        # Restore gatherer state from latest target positions only
        latest_targets = []
        for instrument, targets in restored_state.instrument_to_target_positions.items():
            if targets:  # Only if there are targets for this instrument
                # Get the latest target position (assuming they are sorted by time)
                latest_target = max(targets, key=lambda t: t.time)
                latest_targets.append(latest_target)

        if latest_targets:
            logger.info(f"<yellow>Restoring gatherer state from {len(latest_targets)} latest target positions</yellow>")
            self._position_gathering.restore_from_target_positions(self._context, latest_targets)

    def _handle_warmup_finished(self) -> None:
        if not self._is_data_ready():
            return
        self._strategy.on_warmup_finished(self._context)
        self._context._strategy_state.is_on_warmup_finished_called = True

    def _handle_fit(self, instrument: Instrument | None, event_type: str, data: tuple[dt_64 | None, dt_64]) -> None:
        """
        When scheduled fit event is happened - we need to invoke strategy on_fit method
        """
        if not self._is_data_ready():
            return
        self._fit_is_running = True
        current_time = data[1]
        self._cache.finalize_ohlc_for_instruments(current_time, self._context.instruments)
        self._run_in_thread_pool(self.__invoke_on_fit)

    def _handle_delisting_check(
        self, instrument: Instrument | None, event_type: str, data: tuple[dt_64 | None, dt_64]
    ) -> None:
        """
        Daily delisting check - close positions for instruments delisting within 1 day.
        This is a system-wide scheduled event, so instrument will be None.
        """
        if not self._is_data_ready():
            return

        logger.debug("Performing daily delisting check")

        current_time = data[1]
        current_timestamp = pd.Timestamp(current_time, unit="ns")
        one_day_ahead = current_timestamp + pd.Timedelta(days=1)

        # Find instruments delisting within 1 day
        instruments_to_close = []
        for instr in self._context.instruments:
            if instr.delist_date is not None:
                delist_timestamp = pd.Timestamp(instr.delist_date).tz_localize(None)
                if delist_timestamp <= one_day_ahead:
                    instruments_to_close.append(instr)

        if instruments_to_close:
            logger.info(
                f"Found {len(instruments_to_close)} instruments scheduled for delisting: {instruments_to_close}"
            )

            # Force close positions and remove from universe
            self._universe_manager.remove_instruments(
                instruments_to_close,
                if_has_position_then="close",  # Force close positions
            )

            logger.info("Closed positions and removed instruments scheduled for delisting")

    def _handle_ohlc(self, instrument: Instrument, event_type: str, bar: Bar) -> MarketEvent:
        base_update = self.__update_base_data(instrument, event_type, bar)
        return MarketEvent(self._time_provider.time(), event_type, instrument, bar, is_trigger=base_update)

    def _handle_trade(self, instrument: Instrument, event_type: str, trade: Trade) -> MarketEvent:
        base_update = self.__update_base_data(instrument, event_type, trade)
        return MarketEvent(self._time_provider.time(), event_type, instrument, trade, is_trigger=base_update)

    def _handle_orderbook(self, instrument: Instrument, event_type: str, orderbook: OrderBook) -> MarketEvent:
        base_update = self.__update_base_data(instrument, event_type, orderbook)
        return MarketEvent(self._time_provider.time(), event_type, instrument, orderbook, is_trigger=base_update)

    def _handle_quote(self, instrument: Instrument, event_type: str, quote: Quote) -> MarketEvent:
        base_update = self.__update_base_data(instrument, event_type, quote)
        return MarketEvent(self._time_provider.time(), event_type, instrument, quote, is_trigger=base_update)

    def _handle_funding_payment(
        self, instrument: Instrument, event_type: str, funding_payment: FundingPayment
    ) -> MarketEvent:
        # Apply funding payment to position
        self._account.process_funding_payment(instrument, funding_payment)

        # Continue with existing event processing
        base_update = self.__update_base_data(instrument, event_type, funding_payment)
        return MarketEvent(self._time_provider.time(), event_type, instrument, funding_payment, is_trigger=base_update)

    def _handle_error(self, instrument: Instrument | None, event_type: str, error: BaseErrorEvent) -> None:
        self._strategy.on_error(self._context, error)

    def _handle_order(self, instrument: Instrument, event_type: str, order: Order) -> Order:
        with self._health_monitor("ctx.handle_order"):
            self._account.process_order(order)
            return order

    def _handle_deals(self, instrument: Instrument | None, event_type: str, deals: list[Deal]) -> TriggerEvent | None:
        with self._health_monitor("ctx.handle_deals"):
            if instrument is None:
                logger.debug(
                    f"[<y>{self.__class__.__name__}</y>] :: Execution report for unknown instrument <r>{instrument}</r>"
                )
                return None

            # - process deals only for subscribed instruments
            self._account.process_deals(instrument, deals)
            self._logging.save_deals(instrument, deals)

            # - Process all deals first
            for d in deals:
                # - notify position gatherer and tracker
                self._get_position_gatherer().on_execution_report(self._context, instrument, d)
                self._get_tracker_for(instrument).on_execution_report(self._context, instrument, d)

                logger.debug(
                    f"[<y>{self.__class__.__name__}</y>(<g>{instrument}</g>)] :: executed <r>{d.order_id}</r> | {d.amount} @ {d.price}"
                )

            if self._exporter is not None and (q := self._market_data.quote(instrument)) is not None:
                # - export position changes if exporter is available
                self._exporter.export_position_changes(
                    time=self._time_provider.time(),
                    instrument=instrument,
                    price=q.mid_price(),
                    account=self._account,
                )

            # - notify universe manager about position change
            self._universe_manager.on_alter_position(instrument)

            # - process active targets: if we got 0 position after executions remove current position from active
            if not self._context.get_position(instrument).is_open():
                self._active_targets.pop(instrument, None)

            return None

    def _is_order_update(self, d_type: str) -> bool:
        return d_type in ["order", "deals"]

    def _log_state_mismatch(self) -> None:
        logger.info("<yellow>State comparison between warmup and current state:</yellow>")

        warmup_positions, warmup_orders = self._context.get_warmup_positions(), self._context.get_warmup_orders()

        positions = self._account.get_positions()
        orders = self._account.get_orders()
        instrument_to_orders = defaultdict(list)
        for o in orders.values():
            instrument_to_orders[o.instrument].append(o)

        all_instruments = (
            set(warmup_positions.keys())
            | set(positions.keys())
            | set(warmup_orders.keys())
            | set(instrument_to_orders.keys())
        )

        for instrument in sorted(all_instruments, key=lambda x: x.symbol):
            # Get positions for this instrument
            warmup_pos = warmup_positions.get(instrument)
            current_pos = positions.get(instrument)

            # Get orders for this instrument
            warmup_ord = warmup_orders.get(instrument, [])
            current_ord = instrument_to_orders.get(instrument, [])

            # Format position information
            warmup_pos_info = f"size={warmup_pos.quantity:.6f}" if warmup_pos else "None"
            current_pos_info = f"size={current_pos.quantity:.6f}" if current_pos else "None"

            # Format order information
            warmup_ord_info = f"{len(warmup_ord)} orders" if warmup_ord else "No orders"
            current_ord_info = f"{len(current_ord)} orders" if current_ord else "No orders"

            # Determine if there's a mismatch
            pos_mismatch = (warmup_pos is None) != (current_pos is None) or (
                warmup_pos and current_pos and abs(warmup_pos.quantity - current_pos.quantity) > 1e-10
            )
            ord_mismatch = len(warmup_ord) != len(current_ord)

            # Set color based on mismatch
            color = "<r>" if pos_mismatch or ord_mismatch else "<g>"

            logger.info(
                f"{color}{instrument.symbol}</> - "
                f"Warmup: [Position: {warmup_pos_info}, {warmup_ord_info}] | "
                f"Current: [Position: {current_pos_info}, {current_ord_info}]"
            )

    def get_active_targets(self) -> dict[Instrument, TargetPosition]:
        return self._active_targets

    def emit_signal(self, signal: Signal | list[Signal]) -> None:
        # - add signal to the queue. it will be processed in the data processing loop
        if isinstance(signal, list):
            self._emitted_signals.extend(signal)
        else:
            self._emitted_signals.append(signal)



"""
Main Textual application for running strategies with Jupyter kernel integration.
"""

from pathlib import Path

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.widgets import Footer, Header, Input

from qubx import logger

from .handlers import KernelEventHandler
from .init_code import generate_init_code
from .kernel import IPyKernel
from .widgets import PositionsTable, ReplOutput


class TextualStrategyApp(App[None]):
    """Main Textual application for running strategies with kernel interaction."""

    CSS_PATH = Path(__file__).parent / "styles.tcss"

    BINDINGS = [
        Binding("ctrl+l", "clear_repl", "Clear REPL", show=True),
        Binding("ctrl+c", "interrupt", "Interrupt", show=True),
        Binding("p", "toggle_positions", "Positions", show=True),
        Binding("q", "quit", "Quit", show=True),
    ]

    def __init__(self, config_file: Path, account_file: Path | None, paper: bool, restore: bool) -> None:
        """
        Initialize the Textual strategy app.

        Args:
            config_file: Path to the strategy configuration file
            account_file: Optional path to the account configuration file
            paper: Whether to run in paper trading mode
            restore: Whether to restore the strategy state
        """
        super().__init__()
        self.config_file = config_file
        self.account_file = account_file
        self.paper = paper
        self.restore = restore
        self.kernel = IPyKernel()
        self.output: ReplOutput
        self.input: Input
        self.positions_panel: Vertical
        self.positions_table: PositionsTable
        self.positions_visible = False
        self.event_handler: KernelEventHandler

    async def on_mount(self) -> None:
        """Initialize the app when mounted."""
        # Setup event handler
        self.event_handler = KernelEventHandler(self.output, self.positions_table)

        # Add welcome message
        self.output.write("[bold cyan]Qubx Strategy Runner[/bold cyan]")
        self.output.write("[dim]Type Python commands below and press Enter to execute[/dim]")
        self.output.write("")

        # Start the kernel
        await self.kernel.start()
        self.kernel.register(self.event_handler.handle_event)

        # Initialize the strategy context within the kernel
        self.output.write("[yellow]Initializing strategy context...")
        try:
            # Pre-load the context and helpers into the kernel
            init_code = generate_init_code(
                self.config_file,
                self.account_file,
                self.paper,
                self.restore,
            )
            self.kernel.execute(init_code, silent=False)

            self.output.write("[green]✓ Strategy context initialized and ready!")
            self.output.write("[cyan]Available objects: ctx, S (strategy), portfolio(), orders(), trade(), exit()")
        except Exception as e:
            self.output.write(f"[red]Failed to initialize strategy: {e}")
            logger.exception("Strategy initialization failed")

        # Start interval timer for positions updates (1 second)
        self.set_interval(1.0, self._request_positions)

    async def on_unmount(self) -> None:
        """Clean up when app is closing."""
        # The kernel will handle stopping the context via the exit() function
        await self.kernel.stop()

    def compose(self) -> ComposeResult:
        """Compose the TUI layout."""
        yield Header(show_clock=True)
        with Horizontal(id="main-container"):
            with Vertical(id="output-container"):
                self.output = ReplOutput(id="output", wrap=True, markup=True, max_lines=10000)
                yield self.output
            self.positions_panel = Vertical(id="positions-panel")
            with self.positions_panel:
                self.positions_table = PositionsTable(id="positions-table")
                self.positions_table.setup_columns()
                yield self.positions_table
        with Vertical(id="input-container"):
            self.input = Input(placeholder=">>> Type Python code here and press Enter", id="input")
            yield self.input
        yield Footer()

    # ------------------- Event handlers ---------------
    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle Enter key in the input field."""
        code = event.value.strip()
        if not code:
            return
        self.input.value = ""
        # Echo input
        self.output.write(f"[bold cyan]>>> {code}")
        self.kernel.execute(code)

    # ------------------- Actions ---------------------

    def action_clear_repl(self) -> None:
        """Clear the output."""
        self.output.clear_output()

    async def action_interrupt(self) -> None:
        """Send interrupt signal to the kernel."""
        await self.kernel.interrupt()
        self.output.write("[orange1]⚠ KeyboardInterrupt sent to kernel")

    def action_toggle_positions(self) -> None:
        """Toggle the positions panel visibility."""
        self.positions_visible = not self.positions_visible
        if self.positions_visible:
            self.positions_panel.add_class("visible")
        else:
            self.positions_panel.remove_class("visible")

    # ------------------- Helper methods ----------------

    def _request_positions(self) -> None:
        """Request positions update from the kernel (called by interval timer)."""
        if not self.positions_visible or self.event_handler.is_positions_busy():
            return
        self.event_handler.mark_positions_busy()
        self.kernel.execute("emit_positions()", silent=True)



"""
Base Metric Emitter.

This module provides a base implementation of IMetricEmitter that can be extended by other emitters.
"""

from typing import Any, Dict, List, Optional, Set

import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument, Signal, TargetPosition, dt_64
from qubx.core.interfaces import IAccountViewer, IMetricEmitter, IStrategyContext


class BaseMetricEmitter(IMetricEmitter):
    """
    A base implementation of IMetricEmitter.

    This class provides common functionality for metric emitters, including
    the notify method for time-based emission of metrics.
    """

    # Default set of statistics to emit
    DEFAULT_STATS = {
        "total_capital",
        "net_leverage",
        "gross_leverage",
        "universe_size",
        "price",
        "position_count",
        "position_pnl",
        "position_unrealized_pnl",
        "position_leverage",
    }

    def __init__(
        self, stats_to_emit: Optional[List[str]] = None, stats_interval: str = "1m", tags: dict[str, Any] | None = None
    ):
        """
        Initialize the Base Metric Emitter.

        Args:
            stats_to_emit: Optional list of specific stats to emit. If None, all default stats are emitted.
            stats_interval: Interval for emitting strategy stats (default: "1m")
            tags: Dictionary of default tags/labels to include with all metrics
        """
        self._stats_to_emit: Set[str] = set(stats_to_emit) if stats_to_emit else self.DEFAULT_STATS
        self._stats_interval = pd.Timedelta(stats_interval)
        self._default_tags = tags or {}
        self._last_emission_time = None
        self._context = None

    def _merge_tags(self, tags: dict[str, Any] | None = None, instrument: Instrument | None = None) -> dict[str, Any]:
        """
        Merge default tags with provided tags and instrument tags if provided.
        Also automatically adds is_live tag based on context's simulation state.

        Args:
            tags: Additional tags to merge with default tags
            instrument: Optional instrument to add symbol and exchange tags from

        Returns:
            Dict[str, Any]: Merged tags dictionary
        """
        result = self._default_tags.copy()

        if tags:
            result.update(tags)

        if instrument:
            result.update({"symbol": instrument.symbol, "exchange": instrument.exchange})

        # Add is_live tag based on context's simulation state
        if self._context is not None:
            result["is_live"] = not self._context.is_simulation

        return result

    def _emit_impl(self, name: str, value: float, tags: Dict[str, Any], timestamp: dt_64 | None = None) -> None:
        """
        Implementation of emit to be overridden by subclasses.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Dictionary of tags/labels for the metric (already merged with default tags)
            timestamp: Optional timestamp for the metric
        """
        pass

    def emit(
        self,
        name: str,
        value: float,
        tags: dict[str, Any] | None = None,
        timestamp: dt_64 | None = None,
        instrument: Instrument | None = None,
    ) -> None:
        """
        Emit a metric with the given name, value, and optional tags.

        Args:
            name: Name of the metric
            value: Value of the metric
            tags: Optional dictionary of tags/labels to include with the metric
            timestamp: Optional timestamp for the metric (defaults to current time)
            instrument: Optional instrument to add symbol and exchange tags from
        """
        if self._context is not None and timestamp is None:
            timestamp = self._context.time()
        merged_tags = self._merge_tags(tags, instrument)
        self._emit_impl(name, float(value), merged_tags, timestamp)

    def set_context(self, context: IStrategyContext) -> None:
        """
        Set the strategy context for the metric emitter.

        Args:
            context: The strategy context to use
        """
        self._context = context

    def emit_strategy_stats(self, context: IStrategyContext) -> None:
        """
        Emit standard strategy statistics.

        This method emits standard statistics about the strategy's state, such as
        total capital, leverage, and position information.

        Args:
            context: The strategy context to get statistics from
        """
        # Store context to ensure is_live tag is added
        if self._context is None:
            self._context = context

        try:
            # Get current timestamp
            current_time = context.time()

            tags = {"type": "stats"}

            # Strategy-level metrics
            if "total_capital" in self._stats_to_emit:
                self.emit("total_capital", context.get_total_capital(), tags, timestamp=current_time)

            if "net_leverage" in self._stats_to_emit:
                self.emit("net_leverage", context.get_net_leverage(), tags, timestamp=current_time)

            if "gross_leverage" in self._stats_to_emit:
                self.emit("gross_leverage", context.get_gross_leverage(), tags, timestamp=current_time)

            if "universe_size" in self._stats_to_emit:
                self.emit("universe_size", len(context.instruments), tags, timestamp=current_time)

            if "position_count" in self._stats_to_emit:
                positions = context.get_positions()
                active_positions = [p for i, p in positions.items() if abs(p.quantity) > i.min_size]
                self.emit("position_count", len(active_positions), tags, timestamp=current_time)

            # Position-level metrics
            positions = context.get_positions()
            total_capital = context.get_total_capital()

            for instrument, position in positions.items():
                pos_tags = {"type": "stats"}

                if "price" in self._stats_to_emit and (q := context.quote(instrument)) is not None:
                    self.emit("price", q.mid_price(), pos_tags, timestamp=current_time, instrument=instrument)

                if "position_pnl" in self._stats_to_emit:
                    self.emit("position_pnl", position.pnl, pos_tags, timestamp=current_time, instrument=instrument)

                if "position_unrealized_pnl" in self._stats_to_emit:
                    # Call the method to get the value
                    self.emit(
                        "position_unrealized_pnl",
                        position.unrealized_pnl(),
                        pos_tags,
                        timestamp=current_time,
                        instrument=instrument,
                    )

                if "position_leverage" in self._stats_to_emit and total_capital > 0:
                    position_leverage = context.account.get_leverage(instrument) * 100
                    self.emit(
                        "position_leverage", position_leverage, pos_tags, timestamp=current_time, instrument=instrument
                    )

        except Exception as e:
            logger.error(f"[BaseMetricEmitter] Failed to emit strategy stats: {e}")

    def emit_signals(
        self,
        time: dt_64,
        signals: list["Signal"],
        account: "IAccountViewer",
        target_positions: list["TargetPosition"] | None = None,
    ) -> None:
        """
        Emit signals to the monitoring system.

        Base implementation does nothing - subclasses should override this method
        to implement specific signal emission logic.

        Args:
            time: Timestamp when the signals were generated
            signals: List of signals to emit
            account: Account viewer to get account information like total capital, leverage, etc.
            target_positions: Optional list of target positions generated from the signals
        """
        pass

    def notify(self, context: IStrategyContext) -> None:
        """
        Notify the metric emitter of a time update.

        This method checks if enough time has passed since the last emission
        and emits metrics if necessary.

        Args:
            context: The strategy context to get statistics from
        """
        if not context.is_live_or_warmup:
            return

        # Convert to pandas timestamp for easier time calculations
        current_time = context.time()

        # If this is the first notification, initialize the last emission time
        if self._last_emission_time is None:
            self._last_emission_time = current_time
            return

        # Check if enough time has passed since the last emission
        elapsed = current_time - self._last_emission_time

        if elapsed >= self._stats_interval:
            # logger.debug(f"[{self.__class__.__name__}] Emitting metrics at {current_time}")
            self.emit_strategy_stats(context)
            self._last_emission_time = current_time



from typing import Any

from qubx import logger
from qubx.core.basics import Instrument, MarketType, Order, OrderRequest, OrderSide
from qubx.core.exceptions import OrderNotFound
from qubx.core.interfaces import IAccountProcessor, IBroker, IStrategyContext, ITimeProvider, ITradingManager

from .utils import EXCHANGE_MAPPINGS


class ClientIdStore:
    """Manages generation of unique client order IDs."""

    def __init__(self):
        """Initialize a client ID store."""
        self._order_id: int | None = None

    def generate_id(self, time_provider: ITimeProvider, symbol: str) -> str:
        """Generate a unique client order ID.

        Args:
            time_provider: Time provider to get current timestamp
            symbol: Trading symbol for the order

        Returns:
            A unique client order ID
        """
        # Initialize order ID from timestamp if not yet set
        if self._order_id is None:
            self._order_id = self._initialize_id_from_timestamp(time_provider)

        # Increment order ID to ensure uniqueness across calls
        self._order_id += 1

        # Create and return the unique ID
        return self._create_id(symbol, self._order_id)

    def _initialize_id_from_timestamp(self, time_provider: ITimeProvider) -> int:
        """Initialize the order ID from the current timestamp.

        Args:
            time_provider: Time provider to get current timestamp

        Returns:
            Initial order ID value
        """
        return time_provider.time().astype("int64") // 100_000_000

    def _create_id(self, symbol: str, order_id: int) -> str:
        """Create the ID from symbol and order ID.

        Args:
            symbol: Trading symbol
            order_id: Current order ID counter

        Returns:
            Client ID string
        """
        return "_".join(["qubx", symbol, str(order_id)])


class TradingManager(ITradingManager):
    _context: IStrategyContext
    _brokers: list[IBroker]
    _account: IAccountProcessor
    _strategy_name: str

    _client_id_store: ClientIdStore
    _exchange_to_broker: dict[str, IBroker]

    def __init__(
        self, context: IStrategyContext, brokers: list[IBroker], account: IAccountProcessor, strategy_name: str
    ) -> None:
        self._context = context
        self._brokers = brokers
        self._account = account
        self._strategy_name = strategy_name
        self._client_id_store = ClientIdStore()
        self._exchange_to_broker = {broker.exchange(): broker for broker in brokers}

    def trade(
        self,
        instrument: Instrument,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        client_id: str | None = None,
        **options,
    ) -> Order:
        size_adj = self._adjust_size(instrument, amount)
        side = self._get_side(amount)
        type = self._get_order_type(instrument, price, options)
        price = self._adjust_price(instrument, price, amount)
        client_id = client_id or self._generate_order_client_id(instrument.symbol)

        logger.debug(
            f"[<g>{instrument.symbol}</g>] :: Sending (blocking) {type} {side} {size_adj} {' @ ' + str(price) if price else ''} -> (client_id: <r>{client_id})</r> ..."
        )

        order = self._get_broker(instrument.exchange).send_order(
            instrument=instrument,
            order_side=side,
            order_type=type,
            amount=size_adj,
            price=price,
            time_in_force=time_in_force,
            client_id=client_id,
            **options,
        )

        if order is not None:
            self._account.add_active_orders({order.id: order})

        return order

    def trade_async(
        self,
        instrument: Instrument,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        client_id: str | None = None,
        **options,
    ) -> None:
        size_adj = self._adjust_size(instrument, amount)
        side = self._get_side(amount)
        type = self._get_order_type(instrument, price, options)
        price = self._adjust_price(instrument, price, amount)
        client_id = client_id or self._generate_order_client_id(instrument.symbol)

        logger.debug(
            f"[<g>{instrument.symbol}</g>] :: Sending (async) {type} {side} {size_adj} {' @ ' + str(price) if price else ''} -> (client_id: <r>{client_id})</r> ..."
        )

        self._get_broker(instrument.exchange).send_order_async(
            instrument=instrument,
            order_side=side,
            order_type=type,
            amount=size_adj,
            price=price,
            time_in_force=time_in_force,
            client_id=client_id,
            **options,
        )

    def submit_orders(self, order_requests: list[OrderRequest]) -> list[Order]:
        raise NotImplementedError("Not implemented yet")

    def set_target_position(
        self, instrument: Instrument, target: float, price: float | None = None, time_in_force="gtc", **options
    ) -> Order:
        raise NotImplementedError("Not implemented yet")

    def set_target_leverage(
        self, instrument: Instrument, leverage: float, price: float | None = None, **options
    ) -> Order:
        raise NotImplementedError("Not implemented yet")

    def close_position(self, instrument: Instrument, without_signals: bool = False) -> None:
        position = self._account.get_position(instrument)

        if position.quantity == 0:
            logger.debug(f"[<g>{instrument.symbol}</g>] :: Position already closed or zero size")
            return

        if without_signals:
            closing_amount = -position.quantity
            logger.debug(
                f"[<g>{instrument.symbol}</g>] :: Closing position {position.quantity} with market order for {closing_amount}"
            )
            self.trade(instrument, closing_amount, reduceOnly=True)
        else:
            logger.debug(
                f"[<g>{instrument.symbol}</g>] :: Closing position {position.quantity} by emitting signal with 0 target"
            )
            signal = instrument.signal(self._context, 0, comment="Close position trade")
            self._context.emit_signal(signal)

    def close_positions(self, market_type: MarketType | None = None, without_signals: bool = False) -> None:
        positions = self._account.get_positions()

        positions_to_close = []
        for instrument, position in positions.items():
            if market_type is None or instrument.market_type == market_type:
                if position.is_open():
                    positions_to_close.append(instrument)

        if not positions_to_close:
            logger.debug(f"No open positions to close{f' for market type {market_type}' if market_type else ''}")
            return

        logger.debug(
            f"Closing {len(positions_to_close)} positions{f' for market type {market_type}' if market_type else ''}"
        )

        for instrument in positions_to_close:
            self.close_position(instrument, without_signals)

    def cancel_order(self, order_id: str, exchange: str | None = None) -> bool:
        """Cancel a specific order synchronously."""
        if not order_id:
            return False
        if exchange is None:
            exchange = self._brokers[0].exchange()
        try:
            success = self._get_broker(exchange).cancel_order(order_id)
            if success:
                self._account.remove_order(order_id, exchange)
            return success
        except OrderNotFound:
            # Order was already cancelled or doesn't exist
            # Still try to remove it from account to keep state consistent
            self._account.remove_order(order_id, exchange)
            return False  # Return False since order wasn't found
        except Exception as e:
            logger.error(f"Error canceling order {order_id}: {e}")
            return False  # Return False for any other errors

    def cancel_order_async(self, order_id: str, exchange: str | None = None) -> None:
        """Cancel a specific order asynchronously (non blocking)."""
        if not order_id:
            return
        if exchange is None:
            exchange = self._brokers[0].exchange()
        try:
            self._get_broker(exchange).cancel_order_async(order_id)
            # Note: For async, we remove the order optimistically
            # The actual removal will be confirmed via order status updates
            self._account.remove_order(order_id, exchange)
        except OrderNotFound:
            # Order was already cancelled or doesn't exist
            # Still try to remove it from account to keep state consistent
            self._account.remove_order(order_id, exchange)

    def cancel_orders(self, instrument: Instrument) -> None:
        for o in self._account.get_orders(instrument).values():
            self.cancel_order(o.id, instrument.exchange)

    def update_order(self, order_id: str, price: float, amount: float, exchange: str | None = None) -> Order:
        """Update an existing limit order with new price and amount."""
        if not order_id:
            raise ValueError("Order ID is required")
        if exchange is None:
            exchange = self._brokers[0].exchange()

        # Get the existing order to determine instrument for adjustments
        active_orders = self._account.get_orders(exchange=exchange)
        existing_order = active_orders.get(order_id)
        if not existing_order:
            # Let broker handle the OrderNotFound - just pass through
            logger.debug(f"Updating order {order_id}: {amount} @ {price} on {exchange}")
        else:
            # Apply TradingManager-level adjustments before sending to broker
            instrument = existing_order.instrument
            adjusted_amount = self._adjust_size(instrument, amount)
            adjusted_price = self._adjust_price(instrument, price, amount)
            if adjusted_price is None:
                raise ValueError(f"Price adjustment failed for {instrument.symbol}")
            # Update the values to use adjusted ones
            amount = adjusted_amount
            price = adjusted_price

        try:
            updated_order = self._get_broker(exchange).update_order(order_id, price, abs(amount))

            if updated_order is not None:
                # Update account tracking with new order info
                self._account.process_order(updated_order)
                logger.info(f"[<g>{updated_order.instrument.symbol}</g>] :: Successfully updated order {order_id}")

            return updated_order
        except Exception as e:
            logger.error(f"Error updating order {order_id}: {e}")
            raise e

    def _generate_order_client_id(self, symbol: str) -> str:
        return self._client_id_store.generate_id(self._context, symbol)

    def exchanges(self) -> list[str]:
        return list(self._exchange_to_broker.keys())

    def _adjust_size(self, instrument: Instrument, amount: float) -> float:
        size_adj = instrument.round_size_down(abs(amount))
        if size_adj < instrument.min_size:
            raise ValueError(
                f"[{instrument.symbol}] Attempt to trade size {abs(amount)} less than minimal allowed {instrument.min_size} !"
            )
        return size_adj

    def _adjust_price(self, instrument: Instrument, price: float | None, amount: float) -> float | None:
        if price is None:
            return price
        return instrument.round_price_down(price) if amount > 0 else instrument.round_price_up(price)

    def _get_side(self, amount: float) -> OrderSide:
        return "BUY" if amount > 0 else "SELL"

    def _get_order_type(self, instrument: Instrument, price: float | None, options: dict[str, Any]) -> str:
        if price is None:
            return "market"
        if (stp_type := options.get("stop_type")) is not None:
            return f"stop_{stp_type}"
        return "limit"

    def _get_broker(self, exchange: str) -> IBroker:
        if exchange in self._exchange_to_broker:
            return self._exchange_to_broker[exchange]
        if exchange in EXCHANGE_MAPPINGS and EXCHANGE_MAPPINGS[exchange] in self._exchange_to_broker:
            return self._exchange_to_broker[EXCHANGE_MAPPINGS[exchange]]
        raise ValueError(f"Broker for exchange {exchange} not found")



from qubx import logger
from qubx.core.basics import InitializingSignal, Instrument, Order, Position, TargetPosition
from qubx.core.exceptions import OrderNotFound
from qubx.core.interfaces import IStrategyContext


class StateResolver:
    """
    Collection of static methods for resolving position mismatches between
    warmup simulation and live trading.
    These methods can be used with IStrategyInitializer.set_state_resolver().
    """

    @staticmethod
    def NONE(
        ctx: IStrategyContext,
        sim_positions: dict[Instrument, Position],
        sim_orders: dict[Instrument, list[Order]],
        sim_active_targets: dict[Instrument, TargetPosition],
    ) -> None:
        """
        Do nothing.
        """
        pass

    @staticmethod
    def REDUCE_ONLY(
        ctx: IStrategyContext,
        sim_positions: dict[Instrument, Position],
        sim_orders: dict[Instrument, list[Order]],
        sim_active_targets: dict[Instrument, TargetPosition],
    ) -> None:
        """
        Only allow reducing positions that exist in both simulation and live.

        Args:
            ctx (IStrategyContext): The strategy context
            sim_positions (dict[Instrument, Position]): Positions from the simulation
            sim_orders (dict[Instrument, list[Order]]): Orders from the simulation
        """
        # Get current live positions
        live_positions = ctx.get_positions()

        # Process each live position
        for instrument, live_pos in live_positions.items():
            live_qty = live_pos.quantity

            # Skip positions with zero quantity
            if abs(live_qty) <= instrument.lot_size:
                continue

            # Check if the instrument exists in simulation positions
            if instrument in sim_positions:
                sim_qty = sim_positions[instrument].quantity

                # If signs are opposite, close the live position
                if live_qty * sim_qty < 0:
                    logger.info(f"Closing position for {instrument.symbol} due to opposite direction: {live_qty} -> 0")
                    ctx.emit_signal(InitializingSignal(time=ctx.time(), instrument=instrument, signal=0.0))

                # If live position is larger than sim position (same direction), reduce it
                elif abs(live_qty) > abs(sim_qty) and abs(live_qty) > instrument.lot_size:
                    qty_diff = sim_qty - live_qty
                    logger.info(
                        f"Reducing position for {instrument.symbol}: {live_qty} -> {sim_qty} (diff: {qty_diff:.4f})"
                    )
                    ctx.emit_signal(InitializingSignal(time=ctx.time(), instrument=instrument, signal=sim_qty))

                # If sim position is larger or equal (same direction), do nothing
                else:
                    logger.info(f"Keeping position for {instrument.symbol} as is: {live_qty}")

            # If the instrument doesn't exist in simulation, close the position
            else:
                logger.info(f"Closing position for {instrument.symbol} not in simulation: {live_qty} -> 0")
                ctx.emit_signal(InitializingSignal(time=ctx.time(), instrument=instrument, signal=0.0))

    @staticmethod
    def CLOSE_ALL(
        ctx: IStrategyContext,
        sim_positions: dict[Instrument, Position],
        sim_orders: dict[Instrument, list[Order]],
        sim_active_targets: dict[Instrument, TargetPosition],
    ) -> None:
        """
        Close all positions and start fresh.

        Args:
            ctx (IStrategyContext): The strategy context
            sim_positions (dict[Instrument, Position]): Positions from the simulation
            sim_orders (dict[Instrument, list[Order]]): Orders from the simulation
            sim_active_targets (dict[Instrument, list[TargetPosition]]): Active targets from the simulation
        """
        # TODO: optimize with batch requests
        orders = ctx.get_orders()
        if orders:
            logger.info(f"Cancelling {len(orders)} live orders ...")
            for order in orders.values():
                try:
                    ctx.cancel_order(order.id)
                except OrderNotFound:
                    logger.debug(f"Order {order.id} already cancelled or doesn't exist")

        # Get current live positions
        live_positions = ctx.get_positions()

        # Close all live positions
        for instrument, position in live_positions.items():
            if abs(position.quantity) > instrument.lot_size:
                logger.info(f"Closing position for {instrument.symbol}: {position.quantity} -> 0")
                ctx.emit_signal(InitializingSignal(time=ctx.time(), instrument=instrument, signal=0.0))

    @staticmethod
    def SYNC_STATE(
        ctx: IStrategyContext,
        sim_positions: dict[Instrument, Position],
        sim_orders: dict[Instrument, list[Order]],
        sim_active_targets: dict[Instrument, TargetPosition],
    ) -> None:
        """
        Synchronize the live state with the simulation state.

        Args:
            ctx (IStrategyContext): The strategy context
            sim_positions (dict[Instrument, Position]): Positions from the simulation
            sim_orders (dict[Instrument, list[Order]]): Orders from the simulation
            sim_active_targets (dict[Instrument, list[TargetPosition]]): Active targets from the simulation
        """
        # Get current live positions
        live_positions = ctx.get_positions()

        # - process last active targets from simulation and send them as initializing signals
        for instrument, a_tgt in sim_active_targets.items():
            # - if there is no position in simulation,
            #   but there is a target it means that position is still not open
            #   so we need use limit order
            use_limit_order = instrument in sim_positions and not sim_positions[instrument].is_open()

            s = InitializingSignal(
                time=ctx.time(),
                instrument=instrument,
                signal=a_tgt.target_position_size,
                price=a_tgt.price,
                stop=a_tgt.stop,
                take=a_tgt.take,
                use_limit_order=use_limit_order,
            )
            ctx.emit_signal(s)

        # - now check which positions are open in live and we didn't update them by InitializingSignal
        for instrument, live_pos in live_positions.items():
            if live_pos.is_open() and instrument not in sim_active_targets:
                # - just close the position
                ctx.emit_signal(InitializingSignal(time=ctx.time(), instrument=instrument, signal=0.0))

        # for instrument, sim_pos in sim_positions.items():
        #     live_qty = 0
        #     if instrument in live_positions:
        #         live_qty = live_positions[instrument].quantity

        #     # Calculate the difference needed to match simulation position
        #     qty_diff = sim_pos.quantity - live_qty

        #     # Only trade if there's a difference
        #     if abs(qty_diff) > instrument.lot_size:
        #         logger.info(
        #             f"Syncing position for {instrument.symbol}: {live_qty} -> {sim_pos.quantity} (diff: {qty_diff:.4f})"
        #         )
        #         ctx.trade(instrument, qty_diff)

        # # Close positions that exist in live but not in simulation
        # for instrument, live_pos in live_positions.items():
        #     if instrument not in sim_positions and abs(live_pos.quantity) > instrument.lot_size:
        #         logger.info(f"Closing position for {instrument.symbol} not in simulation: {live_pos.quantity} -> 0")
        #         ctx.trade(instrument, -live_pos.quantity)



import threading
import time
from collections import defaultdict
from dataclasses import dataclass

import numpy as np

from qubx import logger
from qubx.core.basics import CtrlChannel, dt_64
from qubx.core.interfaces import HealthMetrics, IDataArrivalListener, IHealthMonitor, IMetricEmitter, ITimeProvider
from qubx.core.utils import recognize_timeframe
from qubx.utils.collections import DequeFloat64, DequeIndicator


@dataclass
class TimingContext:
    """Context for timing a code block."""

    event_type: str
    start_time: dt_64


class DummyHealthMonitor(IHealthMonitor, IDataArrivalListener):
    """No-op implementation of health metrics monitoring."""

    def __call__(self, event_type: str) -> "DummyHealthMonitor":
        """Support for context manager usage with event type."""
        return self

    def __enter__(self) -> "DummyHealthMonitor":
        """Enter context for timing measurement"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit context and record timing"""
        pass

    def record_event_dropped(self, event_type: str) -> None:
        """Record that an event was dropped."""
        pass

    def on_data_arrival(self, event_type: str, event_time: dt_64) -> None:
        """Record a data arrival time."""
        pass

    def record_start_processing(self, event_type: str, event_time: dt_64) -> None:
        """Record a start processing time."""
        pass

    def record_end_processing(self, event_type: str, event_time: dt_64) -> None:
        """Record a end processing time."""
        pass

    def set_event_queue_size(self, size: int) -> None:
        """Set the current event queue size."""
        pass

    def get_queue_size(self) -> int:
        """Get the current event queue size."""
        return 0

    def get_arrival_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get arrival latency for a specific event type."""
        return 0.0

    def get_queue_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get queue latency for a specific event type."""
        return 0.0

    def get_processing_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get processing latency for a specific event type."""
        return 0.0

    def get_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get end-to-end latency for a specific event type."""
        return 0.0

    def get_execution_latency(self, scope: str, percentile: float = 90) -> float:
        """Get execution latency for a specific scope."""
        return 0.0

    def get_execution_latencies(self) -> dict[str, float]:
        """Get all execution latencies."""
        return {}

    def get_event_frequency(self, event_type: str) -> float:
        """Get the events per second for a specific event type."""
        return 0.0

    def get_system_metrics(self) -> HealthMetrics:
        """Get system-wide metrics."""
        return HealthMetrics(
            queue_size=0.0,
            drop_rate=0.0,
            p50_arrival_latency=0.0,
            p90_arrival_latency=0.0,
            p99_arrival_latency=0.0,
            p50_queue_latency=0.0,
            p90_queue_latency=0.0,
            p99_queue_latency=0.0,
            p50_processing_latency=0.0,
            p90_processing_latency=0.0,
            p99_processing_latency=0.0,
        )

    def start(self) -> None:
        """Start the health metrics monitor."""
        pass

    def stop(self) -> None:
        """Stop the health metrics monitor."""
        pass

    def watch(self, name: str = ""):
        """No-op decorator function that returns the function unchanged.

        This is provided for API compatibility with BaseHealthMonitor.
        """

        def decorator(func):
            return func

        return decorator


class BaseHealthMonitor(IHealthMonitor, IDataArrivalListener):
    """Base implementation of health metrics monitoring using Deque for tracking."""

    def __init__(
        self,
        time_provider: ITimeProvider,
        emitter: IMetricEmitter | None = None,
        emit_interval: str = "1s",
        channel: CtrlChannel | None = None,
        queue_monitor_interval: str = "100ms",
        buffer_size: int = 1000,
        emit_health: bool = True,
    ):
        """Initialize the health metrics monitor.

        Args:
            time_provider: Provider for time information
            emitter: Optional metric emitter for sending metrics to external systems
            emit_interval: Interval to emit metrics, e.g. "1s", "500ms", "5m" (default: "1s")
            channel: Optional data channel to monitor for queue size
            queue_monitor_interval: Interval to check queue size, e.g. "100ms", "500ms" (default: "100ms")
            buffer_size: Size of buffer for storing metrics
            emit_health: Whether to emit health metrics (default: True)
        """
        self.time_provider = time_provider
        self._emitter = emitter
        self._channel = channel
        self._emit_health = emit_health

        # Convert emit interval to nanoseconds
        self._emit_interval_ns = recognize_timeframe(emit_interval)
        self._emit_interval_s = self._emit_interval_ns / 1_000_000_000  # Convert to seconds for sleep

        # Convert queue monitor interval to seconds
        self._queue_monitor_interval_ns = recognize_timeframe(queue_monitor_interval)
        self._queue_monitor_interval_s = self._queue_monitor_interval_ns / 1_000_000_000  # Convert to seconds for sleep

        # Initialize metrics storage
        self._queue_size = DequeFloat64(buffer_size)  # Store last 1000 queue size measurements
        self._event_frequency = defaultdict(lambda: DequeIndicator(buffer_size))
        self._arrival_latency = defaultdict(lambda: DequeIndicator(buffer_size))
        self._start_latency = defaultdict(lambda: DequeIndicator(buffer_size))
        self._end_latency = defaultdict(lambda: DequeIndicator(buffer_size))
        self._dropped_events = defaultdict(lambda: DequeIndicator(buffer_size))
        self._execution_latency = defaultdict(lambda: DequeIndicator(buffer_size))

        # Initialize emission thread control
        self._stop_event = threading.Event()
        self._emission_thread = None

        # Initialize monitor thread control
        self._is_running = False
        self._monitor_thread = None

        # Thread-local storage for timing contexts
        self._thread_local = threading.local()

    def _get_timing_stack(self) -> list[TimingContext]:
        """Get or create the timing stack for the current thread."""
        if not hasattr(self._thread_local, "timing_stack"):
            self._thread_local.timing_stack = []
        return self._thread_local.timing_stack

    def __call__(self, event_type: str) -> "BaseHealthMonitor":
        """Support for context manager usage with event type."""
        self._get_timing_stack().append(TimingContext(event_type, self.time_provider.time()))
        return self

    def __enter__(self) -> "BaseHealthMonitor":
        """Enter context for timing measurement"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit context and record timing"""
        stack = self._get_timing_stack()
        if stack:
            context = stack.pop()
            current_time = self.time_provider.time()
            current_time_ns = current_time.astype("datetime64[ns]").astype(int)
            start_time_ns = context.start_time.astype("datetime64[ns]").astype(int)
            duration = (current_time_ns - start_time_ns) / 1e6  # Convert to ms

            # Record execution latency using event_type as key, not the context object
            self._execution_latency[context.event_type].push_back_fields(current_time_ns, duration)

    def record_event_dropped(self, event_type: str) -> None:
        """Record that an event was dropped."""
        current_time = self.time_provider.time().astype("datetime64[ns]").astype(int)
        self._dropped_events[str(event_type)].push_back_fields(current_time, 1.0)

    def on_data_arrival(self, event_type: str, event_time: dt_64) -> None:
        """Record data arrival time and calculate arrival latency."""
        current_time = self.time_provider.time()
        current_time_ns = current_time.astype("datetime64[ns]").astype(int)
        event_time_ns = event_time.astype("datetime64[ns]").astype(int)
        arrival_latency = (current_time_ns - event_time_ns) / 1e6  # Convert to milliseconds

        # Record arrival latency and event frequency
        self._arrival_latency[str(event_type)].push_back_fields(current_time_ns, arrival_latency)
        self._event_frequency[str(event_type)].push_back_fields(current_time_ns, 1.0)

    def record_start_processing(self, event_type: str, event_time: dt_64) -> None:
        """Record processing start time and calculate queue latency."""
        current_time = self.time_provider.time()
        current_time_ns = current_time.astype("datetime64[ns]").astype(int)
        event_time_ns = event_time.astype("datetime64[ns]").astype(int)
        queue_latency = (current_time_ns - event_time_ns) / 1e6  # Convert to milliseconds

        # Record queue latency
        self._start_latency[str(event_type)].push_back_fields(current_time_ns, queue_latency)

    def record_end_processing(self, event_type: str, event_time: dt_64) -> None:
        """Record processing end time and calculate processing latency."""
        current_time = self.time_provider.time()
        current_time_ns = current_time.astype("datetime64[ns]").astype(int)
        event_time_ns = event_time.astype("datetime64[ns]").astype(int)
        processing_latency = (current_time_ns - event_time_ns) / 1e6  # Convert to milliseconds

        # Record processing latency
        self._end_latency[str(event_type)].push_back_fields(current_time_ns, processing_latency)

    def set_event_queue_size(self, size: int) -> None:
        self._queue_size.push_back(float(size))

    def get_queue_size(self) -> float:
        if self._queue_size.is_empty():
            return 0.0
        return self._queue_size.back()

    def get_arrival_latency(self, event_type: str, percentile: float = 90) -> float:
        return self._get_latency_percentile(event_type, self._arrival_latency, percentile)

    def get_queue_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get queue latency (time from arrival to start processing) for a specific event type."""
        if (
            event_type in self._arrival_latency
            and not self._arrival_latency[event_type].is_empty()
            and event_type in self._start_latency
            and not self._start_latency[event_type].is_empty()
        ):
            arrival_latency = self._get_latency_percentile(event_type, self._arrival_latency, percentile)
            start_latency = self._get_latency_percentile(event_type, self._start_latency, percentile)
            return max(0.0, start_latency - arrival_latency)
        return 0.0

    def get_processing_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get processing latency for a specific event type."""
        if (
            event_type in self._start_latency
            and not self._start_latency[event_type].is_empty()
            and event_type in self._end_latency
            and not self._end_latency[event_type].is_empty()
        ):
            start_latency = self._get_latency_percentile(event_type, self._start_latency, percentile)
            end_latency = self._get_latency_percentile(event_type, self._end_latency, percentile)
            return max(0.0, end_latency - start_latency)
        return 0.0

    def get_latency(self, event_type: str, percentile: float = 90) -> float:
        """Get the end-to-end latency for a specific event type (for backwards compatibility)."""
        if event_type not in self._end_latency or self._end_latency[str(event_type)].is_empty():
            return 0.0
        latencies = self._end_latency[str(event_type)].to_array()["value"]
        return float(np.percentile(latencies, percentile))

    def get_execution_latency(self, scope: str, percentile: float = 90) -> float:
        return self._get_latency_percentile(scope, self._execution_latency, percentile)

    def get_execution_latencies(self) -> dict[str, float]:
        scopes = self._execution_latency.keys()
        return {scope: self.get_execution_latency(scope) for scope in scopes}

    def get_event_frequency(self, event_type: str) -> float:
        """Get the events per second for a specific event type."""
        if event_type not in self._event_frequency or self._event_frequency[event_type].is_empty():
            return 0.0

        series = self._event_frequency[event_type].to_array()
        if len(series) < 2:  # Need at least 2 points to calculate frequency
            return float(np.sum(series["value"]))  # Return total events if only one point

        # Calculate time window between newest and oldest events
        newest_time = series[-1]["timestamp"]  # Last event in the array
        oldest_time = series[0]["timestamp"]  # First event in the array
        time_window_ns = newest_time - oldest_time

        # Avoid division by zero and ensure minimum window
        if time_window_ns <= 0:
            return float(np.sum(series["value"]))  # Return total events if all events at same time

        # Convert window to seconds for per-second frequency
        time_window_seconds = max(1.0, time_window_ns / 1e9)  # At least 1 second window

        # Sum all events in our window
        total_events = np.sum(series["value"])

        # Calculate events per second
        return float(total_events / time_window_seconds)

    def get_system_metrics(self) -> HealthMetrics:
        """Get aggregated system metrics."""
        # Calculate average queue size
        avg_queue_size = float(np.mean(self._queue_size.to_array())) if not self._queue_size.is_empty() else 0.0

        # Calculate dropped events rate
        drop_rate = self._calc_total_drop_rate(self._dropped_events)

        # Calculate latency percentiles
        p50_arrival_latency = p90_arrival_latency = p99_arrival_latency = 0.0
        p50_start_latency = p90_start_latency = p99_start_latency = 0.0
        p50_end_latency = p90_end_latency = p99_end_latency = 0.0
        p50_queue_latency = p90_queue_latency = p99_queue_latency = 0.0
        p50_processing_latency = p90_processing_latency = p99_processing_latency = 0.0

        # Aggregate latencies across all event types
        arrival_latencies = []
        start_latencies = []
        end_latencies = []

        event_types = self._arrival_latency.keys()
        for event_type in event_types:
            if not self._arrival_latency[event_type].is_empty():
                arrival_latencies.extend(self._arrival_latency[event_type].to_array()["value"])
            if not self._start_latency[event_type].is_empty():
                start_latencies.extend(self._start_latency[event_type].to_array()["value"])
            if not self._end_latency[event_type].is_empty():
                end_latencies.extend(self._end_latency[event_type].to_array()["value"])

        if arrival_latencies:
            p50_arrival_latency = float(np.percentile(arrival_latencies, 50))
            p90_arrival_latency = float(np.percentile(arrival_latencies, 90))
            p99_arrival_latency = float(np.percentile(arrival_latencies, 99))

        if start_latencies:
            p50_start_latency = float(np.percentile(start_latencies, 50))
            p90_start_latency = float(np.percentile(start_latencies, 90))
            p99_start_latency = float(np.percentile(start_latencies, 99))
            p50_queue_latency = p50_start_latency - p50_arrival_latency
            p90_queue_latency = p90_start_latency - p90_arrival_latency
            p99_queue_latency = p99_start_latency - p99_arrival_latency

        if end_latencies:
            p50_end_latency = float(np.percentile(end_latencies, 50))
            p90_end_latency = float(np.percentile(end_latencies, 90))
            p99_end_latency = float(np.percentile(end_latencies, 99))
            p50_processing_latency = p50_end_latency - p50_start_latency
            p90_processing_latency = p90_end_latency - p90_start_latency
            p99_processing_latency = p99_end_latency - p99_start_latency

        return HealthMetrics(
            queue_size=avg_queue_size,
            drop_rate=drop_rate,
            p50_arrival_latency=p50_arrival_latency,
            p90_arrival_latency=p90_arrival_latency,
            p99_arrival_latency=p99_arrival_latency,
            p50_queue_latency=p50_queue_latency,
            p90_queue_latency=p90_queue_latency,
            p99_queue_latency=p99_queue_latency,
            p50_processing_latency=p50_processing_latency,
            p90_processing_latency=p90_processing_latency,
            p99_processing_latency=p99_processing_latency,
        )

    def start(self) -> None:
        """Start the metrics emission thread and queue monitoring thread."""
        # Start queue size monitoring if channel is provided
        if self._channel is not None:
            self._is_running = True
            self._monitor_thread = threading.Thread(target=self._monitor_queue_size, daemon=True)
            self._monitor_thread.start()

        # Start metrics emission if emitter is provided
        if self._emitter is not None:

            def emit_metrics():
                while not self._stop_event.is_set():
                    try:
                        self._emit()
                    except Exception as e:
                        logger.error(f"Error emitting metrics: {e}")
                    finally:
                        time.sleep(self._emit_interval_s)

            self._stop_event.clear()
            self._emission_thread = threading.Thread(target=emit_metrics, daemon=True)
            self._emission_thread.start()

    def stop(self) -> None:
        """Stop the metrics emission thread and queue monitoring thread."""

        # Stop queue size monitoring
        if self._monitor_thread is not None:
            self._is_running = False
            self._monitor_thread.join()
            self._monitor_thread = None

        # Stop metrics emission
        if self._emission_thread is not None:
            self._stop_event.set()
            self._emission_thread.join()
            self._emission_thread = None

    def _monitor_queue_size(self) -> None:
        """Background thread for monitoring queue size."""
        while self._is_running:
            try:
                # Update queue size if we have a channel
                if self._channel is not None:
                    current_size = self._channel._queue.qsize()
                    self.set_event_queue_size(current_size)
            except Exception as e:
                logger.error(f"Error monitoring queue size: {e}")
            finally:
                time.sleep(self._queue_monitor_interval_s)

    def _get_latency_percentile(self, event_type: str, latencies: dict, percentile: float) -> float:
        if event_type not in latencies or latencies[event_type].is_empty():
            return 0.0
        _latencies = latencies[event_type].to_array()["value"]
        return float(np.percentile(_latencies, percentile))

    def _emit(self) -> None:
        """Emit all metrics to the configured emitter."""
        if not self._emit_health or self._emitter is None:
            return

        metrics = self.get_system_metrics()
        current_time = self.time_provider.time()
        tags = {"type": "health"}

        # Emit system-wide metrics
        self._emitter.emit(name="health.queue_size", value=metrics.queue_size, tags=tags, timestamp=current_time)
        self._emitter.emit(
            name="health.dropped_events",
            value=metrics.drop_rate,
            tags=tags,
            timestamp=current_time,
        )

        # Emit latency metrics with percentiles
        self._emitter.emit(
            name="health.arrival_latency.p50",
            value=metrics.p50_arrival_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.arrival_latency.p90",
            value=metrics.p90_arrival_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.arrival_latency.p99",
            value=metrics.p99_arrival_latency,
            tags=tags,
            timestamp=current_time,
        )

        self._emitter.emit(
            name="health.queue_latency.p50",
            value=metrics.p50_queue_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.queue_latency.p90",
            value=metrics.p90_queue_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.queue_latency.p99",
            value=metrics.p99_queue_latency,
            tags=tags,
            timestamp=current_time,
        )

        self._emitter.emit(
            name="health.processing_latency.p50",
            value=metrics.p50_processing_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.processing_latency.p90",
            value=metrics.p90_processing_latency,
            tags=tags,
            timestamp=current_time,
        )
        self._emitter.emit(
            name="health.processing_latency.p99",
            value=metrics.p99_processing_latency,
            tags=tags,
            timestamp=current_time,
        )

        # Collect all unique event types from all metric dictionaries
        event_types = set()
        event_types.update(self._dropped_events.keys())
        event_types.update(self._arrival_latency.keys())
        event_types.update(self._start_latency.keys())
        event_types.update(self._end_latency.keys())
        event_types.update(self._event_frequency.keys())

        # Emit per-event metrics for all event types
        for event_type in event_types:
            freq = self.get_event_frequency(event_type)
            processing_latency = self.get_processing_latency(event_type)
            drop_rate = self._get_drop_rate(event_type)
            arrival_latency = self.get_arrival_latency(event_type)
            queue_latency = self.get_queue_latency(event_type)

            event_tags = {"type": "health", "event_type": str(event_type)}
            self._emitter.emit("health.event_frequency", freq, event_tags, current_time)
            self._emitter.emit("health.event_processing_latency", processing_latency, event_tags, current_time)
            self._emitter.emit("health.event_drop_rate", drop_rate, event_tags, current_time)
            self._emitter.emit("health.event_arrival_latency", arrival_latency, event_tags, current_time)
            self._emitter.emit("health.event_queue_latency", queue_latency, event_tags, current_time)

        for scope, latency in self.get_execution_latencies().items():
            self._emitter.emit("health.execution_latency", latency, {"type": "health", "scope": scope}, current_time)

    def _calc_weighted_latency_avg(self, latency_dict) -> tuple[float, float, float]:
        """Calculate weighted average latency and percentiles across all event types.

        Returns:
            Tuple of (weighted_average, p90, p99) latencies in milliseconds.
        """
        if not latency_dict:
            return 0.0, 0.0, 0.0

        all_latencies = []
        all_timestamps = []

        for series in latency_dict.values():
            if not series.is_empty():
                arr = series.to_array()
                all_latencies.extend(arr["value"])
                all_timestamps.extend(arr["timestamp"])

        if not all_latencies:
            return 0.0, 0.0, 0.0

        # Convert to numpy arrays for efficient computation
        latencies = np.array(all_latencies)
        timestamps = np.array(all_timestamps)

        # Sort by timestamp to maintain temporal relationship
        sort_idx = np.argsort(timestamps)
        latencies = latencies[sort_idx]

        # Calculate statistics
        avg = float(np.mean(latencies))
        p90 = float(np.percentile(latencies, 90))
        p99 = float(np.percentile(latencies, 99))

        return avg, p90, p99

    def _get_drop_rate(self, event_type: str) -> float:
        """Calculate the drop rate for a specific event type."""
        if event_type not in self._dropped_events or self._dropped_events[event_type].is_empty():
            return 0.0

        dropped_data = self._dropped_events[event_type].to_array()
        if len(dropped_data) < 2:
            # If we only have one data point, return the total number of drops
            return float(np.sum(dropped_data["value"]))

        # Calculate time window between newest and oldest drops
        newest_time = dropped_data[-1]["timestamp"]
        oldest_time = dropped_data[0]["timestamp"]
        time_window_ns = newest_time - oldest_time

        # Avoid division by zero and ensure minimum window
        if time_window_ns <= 0:
            return float(np.sum(dropped_data["value"]))

        # Convert window to seconds for per-second rate
        time_window_seconds = max(1.0, time_window_ns / 1e9)  # At least 1 second window

        # Calculate drops per second
        total_drops = np.sum(dropped_data["value"])
        return float(total_drops / time_window_seconds)

    def _calc_total_drop_rate(self, dropped_dict) -> float:
        """Calculate the rate of dropped events per second across all event types."""
        if not dropped_dict:
            return 0.0

        total_drop_rate = 0
        total_time_s = 0.0

        for event_type, series in dropped_dict.items():
            # Get drop rate for each event type
            drop_rate = self._get_drop_rate(event_type)
            total_drop_rate += drop_rate
            total_time_s += 1.0  # Each event type contributes 1 second to normalize

        if total_time_s <= 0:
            return 0.0

        # Return average drop rate across all event types
        return float(total_drop_rate / total_time_s)

    def watch(self, scope_name: str = ""):
        """Decorator function to time a function execution using health monitor.

        Args:
            scope_name: Name for the timing scope. If empty string is provided,
                       function's qualified name will be used.

        Returns:
            Decorator function that times the decorated function.

        Example:
            @health_monitor.watch()
            def my_function():
                # This function execution will be timed using function's qualified name
                ...

            @health_monitor.watch("custom_scope")
            def another_function():
                # This function will be timed under the name "custom_scope"
                ...
        """

        def decorator(func):
            nonlocal scope_name
            # Use function's qualified name if scope_name is empty
            if scope_name == "":
                scope_name = f"{func.__module__}.{func.__qualname__}"

            def wrapper(*args, **kwargs):
                with self(scope_name):
                    return func(*args, **kwargs)

            # Preserve function metadata
            wrapper.__name__ = func.__name__
            wrapper.__doc__ = func.__doc__
            wrapper.__module__ = func.__module__
            wrapper.__qualname__ = func.__qualname__
            wrapper.__annotations__ = func.__annotations__

            return wrapper

        return decorator



import asyncio
import re
from collections import defaultdict
from typing import Any, Awaitable, Callable, Dict, List, Set

import numpy as np
import pandas as pd

import ccxt.pro as cxp
from ccxt import BadSymbol
from qubx import logger
from qubx.core.basics import (
    AssetBalance,
    Deal,
    FundingRate,
    Instrument,
    Liquidation,
    OpenInterest,
    Order,
    Position,
    dt_64,
)
from qubx.core.series import OrderBook, Quote, Trade
from qubx.core.utils import recognize_time
from qubx.utils.marketdata.ccxt import (
    ccxt_symbol_to_instrument,
)
from qubx.utils.orderbook import accumulate_orderbook_levels
from qubx.utils.time import to_utc_naive

from .exceptions import (
    CcxtLiquidationParsingError,
    CcxtSymbolNotRecognized,
)

EXCH_SYMBOL_PATTERN = re.compile(r"(?P<base>[^/]+)/(?P<quote>[^:]+)(?::(?P<margin>.+))?")


def ccxt_convert_order_info(instrument: Instrument, raw: dict[str, Any]) -> Order:
    """
    Convert CCXT excution record to Order object
    """
    ri = raw["info"]
    if isinstance(ri, list):
        # we don't handle case when order info is a list
        ri = {}
    # Prioritize outer level (exchange-specific parsing) for amount
    amnt_raw = raw.get("amount") or ri.get("origQty")
    if amnt_raw is None:
        # Try alternative fields for different exchanges
        amnt_raw = ri.get("sz") or ri.get("origSz") or 0.0
    amnt = float(amnt_raw)
    price = raw["price"] or 0.0
    status = raw["status"] or "UNKNOWN"
    side_raw = raw["side"]
    if side_raw is None:
        side = "UNKNOWN"
    else:
        side = side_raw.upper()
    # Prioritize outer level (exchange-specific parsing overrides) over inner level
    _type = raw.get("type", ri.get("type"))
    if _type is None:
        # Fallback for exchanges that don't provide type info
        _type = "UNKNOWN"
    else:
        _type = _type.upper()
    if status == "open":
        status = ri.get("status", status)  # for filled / part_filled ?

    # Ensure status is always a string and uppercase
    if not status:
        status = "UNKNOWN"

    status = status.upper()
    options = {}
    if raw.get("reduceOnly"):
        options["reduceOnly"] = True

    tif = raw.get("timeInForce")

    return Order(
        id=raw["id"],
        type=_type,
        instrument=instrument,
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        quantity=abs(amnt) * (-1 if side == "SELL" else 1),
        price=float(price) if price is not None else 0.0,
        side=side,
        status=status,
        time_in_force=tif,
        client_id=raw["clientOrderId"],
        cost=float(raw["cost"] or 0),  # cost can be None
        options=options,
    )


def ccxt_convert_deal_info(raw: Dict[str, Any]) -> Deal:
    fee_amount = None
    fee_currency = None
    if "fee" in raw:
        fee_amount = float(raw["fee"]["cost"])
        fee_currency = raw["fee"]["currency"]
    return Deal(
        id=raw["id"],
        order_id=raw["order"],
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        amount=float(raw["amount"]) * (-1 if raw["side"] == "sell" else +1),
        price=float(raw["price"]),
        aggressive=raw["takerOrMaker"] == "taker",
        fee_amount=fee_amount,
        fee_currency=fee_currency,
    )


def ccxt_extract_deals_from_exec(report: Dict[str, Any]) -> List[Deal]:
    """
    Small helper for extracting deals (trades) from CCXT execution report
    """
    deals = list()
    if trades := report.get("trades"):
        for t in trades:
            deals.append(ccxt_convert_deal_info(t))
    return deals


def ccxt_restore_position_from_deals(
    pos: Position, current_volume: float, deals: List[Deal], reserved_amount: float = 0.0
) -> Position:
    if current_volume != 0:
        instr = pos.instrument
        _last_deals = []

        # - try to find last deals that led to this position
        for d in sorted(deals, key=lambda x: x.time, reverse=True):
            current_volume -= d.amount
            # - spot case when fees may be deducted from the base coin
            #   that may decrease total amount
            if d.fee_amount is not None:
                if instr.base == d.fee_currency:
                    current_volume += d.fee_amount
            # print(d.amount, current_volume)
            _last_deals.insert(0, d)

            # - take in account reserves
            if abs(current_volume) - abs(reserved_amount) < instr.lot_size:
                break

        # - reset to 0
        pos.reset()

        if abs(current_volume) - abs(reserved_amount) > instr.lot_size:
            # - - - TODO - - - !!!!
            logger.warning(
                f"Couldn't restore full deals history for {instr.symbol} symbol. Qubx will use zero position !"
            )
        else:
            fees_in_base = 0.0
            for d in _last_deals:
                pos.update_position_by_deal(d)
                if d.fee_amount is not None:
                    if instr.base == d.fee_currency:
                        fees_in_base += d.fee_amount
            # - we round fees up in case of fees in base currency
            pos.quantity -= pos.instrument.round_size_up(fees_in_base)
    return pos


def ccxt_convert_trade(trade: dict[str, Any]) -> Trade:
    t_ns = trade["timestamp"] * 1_000_000  # this is trade time
    price, amnt = trade["price"], trade["amount"]
    side = int(trade["side"] == "buy") * 2 - 1
    return Trade(t_ns, price, amnt, side)


def ccxt_convert_positions(
    pos_infos: list[dict], ccxt_exchange_name: str, markets: dict[str, dict[str, Any]]
) -> list[Position]:
    positions = []
    for info in pos_infos:
        symbol = info["symbol"]
        if symbol not in markets:
            logger.warning(f"Could not find symbol {symbol}, skipping position...")
            continue
        instr = ccxt_symbol_to_instrument(
            ccxt_exchange_name,
            markets[symbol],
        )
        pos = Position(
            instrument=instr,
            quantity=abs(info["contracts"]) * (-1 if info["side"] == "short" else 1),
            pos_average_price=info["entryPrice"],
        )
        if info.get("markPrice", None) is not None:
            pos.update_market_price(pd.Timestamp(info["timestamp"], unit="ms").asm8, info["markPrice"], 1)
        positions.append(pos)
    return positions


def ccxt_convert_orderbook(
    ob: dict,
    instr: Instrument,
    levels: int = 50,
    tick_size_pct: float = 0.01,
    sizes_in_quoted: bool = False,
    current_timestamp: dt_64 | None = None,
) -> OrderBook | None:
    """
    Convert a ccxt order book to an OrderBook object with a fixed tick size.

    Parameters:
        ob (dict): The order book dictionary from ccxt.
        instr (Instrument): The instrument object containing market-specific details.
        levels (int, optional): The number of levels to include in the order book. Default is 50.
        tick_size_pct (float, optional): The tick size percentage. Default is 0.01%.
        sizes_in_quoted (bool, optional): Whether the size is in the quoted currency. Default is False.

    Returns:
        OrderBook: The converted OrderBook object.
    """
    try:
        # Convert timestamp to nanoseconds as a long long integer
        dt = recognize_time(ob["datetime"]) if ob["datetime"] is not None else current_timestamp

        # Determine tick size
        if tick_size_pct == 0:
            tick_size = instr.tick_size
        else:
            # Calculate mid price from the top of the book
            top_bid = ob["bids"][0][0] if ob["bids"] else 0
            top_ask = ob["asks"][0][0] if ob["asks"] else 0

            if top_bid == 0 or top_ask == 0:
                # If either is missing, use the other one
                mid_price = top_bid or top_ask
            else:
                mid_price = (top_bid + top_ask) / 2

            # Calculate tick size as percentage of mid price
            raw_tick_size = max(mid_price * tick_size_pct / 100, instr.tick_size)

            # Round down tick_size to align with instrument's minimum tick size
            tick_size = instr.round_price_down(raw_tick_size)

        # Pre-allocate buffers for bids and asks
        bids_buffer = np.zeros(levels, dtype=np.float64)
        asks_buffer = np.zeros(levels, dtype=np.float64)

        raw_bids = np.array(ob["bids"])
        raw_asks = np.array(ob["asks"])

        # Extract price and size columns from raw bids and asks
        # Some exchanges return more than 2 columns for bids and asks
        raw_bids = raw_bids[:, :2].astype(np.float64)
        raw_asks = raw_asks[:, :2].astype(np.float64)

        # Accumulate bids and asks into the buffers
        top_bid, bids = accumulate_orderbook_levels(raw_bids, bids_buffer, tick_size, True, levels, sizes_in_quoted)

        top_ask, asks = accumulate_orderbook_levels(raw_asks, asks_buffer, tick_size, False, levels, sizes_in_quoted)

        # Create and return the OrderBook object
        return OrderBook(
            time=dt,
            top_bid=top_bid,
            top_ask=top_ask,
            tick_size=tick_size,
            bids=bids,
            asks=asks,
        )
    except Exception as e:
        from pprint import pformat

        logger.error(f"Failed to convert order book for {instr}: {e}")
        logger.error(pformat(ob))
        return None


def ccxt_convert_liquidation(liq: dict[str, Any]) -> Liquidation:
    try:
        _dt = to_utc_naive(pd.Timestamp(liq["datetime"])).asm8
        return Liquidation(
            time=_dt,
            price=liq["price"],
            quantity=liq["contracts"],
            side=(1 if liq["info"]["S"] == "BUY" else -1),
        )
    except Exception as e:
        raise CcxtLiquidationParsingError(f"Failed to parse liquidation: {e}")


def ccxt_convert_ticker(ticker: dict[str, Any]) -> Quote:
    """
    Convert a ccxt ticker to a Quote object.
    Parameters:
        ticker (dict): The ticker dictionary from ccxt.
        instr (Instrument): The instrument object containing market-specific details.
    Returns:
        Quote: The converted Quote object.
    """
    return Quote(
        time=to_utc_naive(pd.Timestamp(ticker["datetime"])).asm8,
        bid=ticker["bid"],
        ask=ticker["ask"],
        bid_size=ticker["bidVolume"],
        ask_size=ticker["askVolume"],
    )


def ccxt_convert_funding_rate(info: dict[str, Any]) -> FundingRate:
    return FundingRate(
        time=pd.Timestamp(info["timestamp"], unit="ms").asm8,
        rate=info["fundingRate"],
        interval=info["interval"],
        next_funding_time=pd.Timestamp(info["nextFundingTime"], unit="ms").asm8,
        mark_price=info.get("markPrice"),
        index_price=info.get("indexPrice"),
    )


def ccxt_convert_balance(d: dict[str, Any]) -> dict[str, AssetBalance]:
    balances = {}
    for currency, data in d["total"].items():
        if not data:
            continue
        total = float(d["total"].get(currency, 0) or 0)
        locked = float(d["used"].get(currency, 0) or 0)
        balances[currency] = AssetBalance(free=total - locked, locked=locked, total=total)
    return balances


def ccxt_convert_open_interest(symbol: str, info: dict[str, Any]) -> OpenInterest:
    # Extract open interest amount (base asset amount)
    open_interest_amount = info.get("openInterestAmount", 0.0)

    # Try to get USD value from multiple possible fields
    open_interest_usd = info.get("openInterestValue") or info.get("openInterestUsd") or info.get("notional") or 0.0

    # If USD value is still 0 or None, we'll need to calculate it using mark price
    # For now, we'll use 0.0 and let the calling code handle price conversion if needed
    if open_interest_usd is None:
        open_interest_usd = 0.0

    # Handle timestamp conversion more robustly
    timestamp = info.get("timestamp")
    if timestamp is None:
        raise ValueError("Missing timestamp in open interest data")

    # Convert timestamp to dt_64 format more robustly
    try:
        # First try as milliseconds (most common CCXT format)
        time_dt = pd.Timestamp(timestamp, unit="ms").asm8
    except (ValueError, TypeError, pd.errors.OutOfBoundsDatetime) as e:
        try:
            # Try as pandas timestamp without unit specification
            time_dt = pd.Timestamp(timestamp).asm8
        except (ValueError, TypeError, pd.errors.OutOfBoundsDatetime) as e2:
            try:
                # Try parsing as datetime string and convert to UTC naive
                time_dt = to_utc_naive(pd.Timestamp(timestamp)).asm8
            except Exception as e3:
                # Include detailed information about the timestamp that failed
                from qubx import logger

                logger.error(
                    f"Open interest timestamp conversion failed - value: {timestamp}, type: {type(timestamp)}, repr: {repr(timestamp)}"
                )
                logger.error(f"Method 1 (ms unit) error: {e}")
                logger.error(f"Method 2 (no unit) error: {e2}")
                logger.error(f"Method 3 (utc_naive) error: {e3}")
                raise ValueError(
                    f"Could not convert timestamp {timestamp} (type: {type(timestamp)}) to datetime. Original error: {e}"
                ) from e

    return OpenInterest(
        time=time_dt,
        symbol=symbol,
        open_interest=float(open_interest_amount),
        open_interest_usd=float(open_interest_usd),
    )


def find_instrument_for_exch_symbol(exch_symbol: str, symbol_to_instrument: Dict[str, Instrument]) -> Instrument:
    match = EXCH_SYMBOL_PATTERN.match(exch_symbol)
    if not match:
        raise CcxtSymbolNotRecognized(f"Invalid exchange symbol {exch_symbol}")
    base = match.group("base")
    quote = match.group("quote")
    symbol = f"{base}{quote}"
    if symbol not in symbol_to_instrument:
        raise CcxtSymbolNotRecognized(f"Unknown symbol {symbol}")
    return symbol_to_instrument[symbol]


def instrument_to_ccxt_symbol(instr: Instrument) -> str:
    return f"{instr.base}/{instr.quote}:{instr.settle}" if instr.is_futures() else f"{instr.base}/{instr.quote}"


def ccxt_find_instrument(
    symbol: str, exchange: cxp.Exchange, symbol_to_instrument: Dict[str, Instrument] | None = None
) -> Instrument:
    instrument = None
    if symbol_to_instrument is not None:
        instrument = symbol_to_instrument.get(symbol)
        if instrument is not None:
            return instrument
        try:
            instrument = find_instrument_for_exch_symbol(symbol, symbol_to_instrument)
        except CcxtSymbolNotRecognized:
            pass
    if instrument is None:
        try:
            symbol_info = exchange.market(symbol)
        except BadSymbol:
            raise CcxtSymbolNotRecognized(f"Unknown symbol {symbol}")
        exchange_name = exchange.name
        assert exchange_name is not None
        instrument = ccxt_symbol_to_instrument(exchange_name, symbol_info)
    if symbol_to_instrument is not None and symbol not in symbol_to_instrument:
        symbol_to_instrument[symbol] = instrument
    return instrument


def create_market_type_batched_subscriber(
    subscriber: Callable[[List[Instrument]], Awaitable[None]], instruments: Set[Instrument]
) -> Callable[[], Awaitable[None]]:
    """
    Create a batched subscriber that calls the original subscriber for each market type group.

    This utility function groups instruments by market type and calls the subscriber function
    for each group separately. This is necessary because some exchanges require separate
    calls for different market types (e.g., spot vs futures).

    Args:
        subscriber: Function to call for each market type group
        instruments: Set of instruments to group by market type

    Returns:
        Async function that will call subscriber for each market type group
    """
    # Group instruments by market type
    instr_by_type: Dict[str, List[Instrument]] = defaultdict(list)
    for instr in instruments:
        instr_by_type[instr.market_type].append(instr)

    # Sort instruments by symbol within each group for consistent ordering
    for instrs in instr_by_type.values():
        instrs.sort(key=lambda i: i.symbol)

    async def batched_subscriber():
        """Execute subscriber for each market type group concurrently."""
        await asyncio.gather(*[subscriber(instrs) for instrs in instr_by_type.values()])

    return batched_subscriber
